import shutil
import subprocess
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import remove_build_directory


class Test(TestCase):

    def run_test_assert(self, name, expected):
        proc = subprocess.run(['./build/debug/test', name],
                              capture_output=True,
                              text=True)
        self.assertNotEqual(proc.returncode, 0)
        self.assert_in(expected, proc.stderr + proc.stdout)

    def run_app_assert(self, expected, path, not_expected=None):
        proc = subprocess.run([f'./build/{path}/app'],
                              capture_output=True,
                              text=True)
        self.assertNotEqual(proc.returncode, 0)
        self.assert_in(expected, proc.stderr)

        if not_expected is not None:
            self.assert_not_in(not_expected, proc.stderr + proc.stdout)

    def test_traceback(self):
        name = 'test_traceback'
        remove_build_directory(name)

        shutil.copytree('tests/files/traceback', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            try:
                with patch('sys.argv', ['mys', 'test', '-v']):
                    mys.cli.main()
            except SystemExit:
                pass

            self.run_test_assert(
                'test_panic',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 5 in test_panic\n'
                '    print(""[1])\n'
                '\n'
                'Panic(message="String index 1 is out of range.")\n')

            self.run_test_assert(
                'test_panic_2',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 13 in test_panic_2\n'
                '    panic_2()\n'
                '  File: "./src/lib.mys", line 9 in panic_2\n'
                '    print(""[i])\n'
                '\n'
                'Panic(message="String index 10 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_except',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 23 in test_panic_in_except\n'
                '    print(b""[11])\n'
                '\n'
                'Panic(message="Bytes index 11 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_if',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 28 in test_panic_in_if\n'
                '    print(b""[5])\n'
                '\n'
                'Panic(message="Bytes index 5 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_else',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 35 in test_panic_in_else\n'
                '    print(b""[6])\n'
                '\n'
                'Panic(message="Bytes index 6 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_for',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 40 in test_panic_in_for\n'
                '    print(b"123"[i])\n'
                '\n'
                'Panic(message="Bytes index 3 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_while',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 45 in test_panic_in_while\n'
                '    print(b""[10])\n'
                '\n'
                'Panic(message="Bytes index 10 is out of range.")\n')

            self.run_test_assert(
                'test_panic_in_match',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 51 in test_panic_in_match\n'
                '    print(b""[-1])\n'
                '\n'
                'Panic(message="Bytes index -1 is out of range.")\n')

            self.run_test_assert(
                'test_error_in_runtime',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 59 in test_error_in_runtime\n'
                '    raise MyError(1, 2)\n'
                '\n'
                'MyError(x=1, y=2)\n')

            self.run_test_assert(
                'test_error_in_fiber',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 64 in run\n'
                '    raise MyError(2, 3)\n'
                '\n'
                'MyError(x=2, y=3)\n')

            self.run_test_assert(
                'test_modulo_zero',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 78 in test_modulo_zero\n'
                '    assert modulo(10, 0) == 0\n'
                '  File: "./src/lib.mys", line 73 in modulo\n'
                '    return a % b\n'
                '\n'
                'ValueError(message="cannot divide or modulo by zero")\n')

            self.run_test_assert(
                'test_division_by_zero',
                'Traceback (most recent call last):\n'
                '  File: "./src/lib.mys", line 86 in test_division_by_zero\n'
                '    assert divide(10, 0) == 5\n'
                '  File: "./src/lib.mys", line 81 in divide\n'
                '    return a / b\n'
                '\n'
                'ValueError(message="cannot divide or modulo by zero")\n')

            # Debug build.
            with patch('sys.argv', ['mys', 'build', '-o', 'debug']):
                mys.cli.main()

            self.run_app_assert(
                'Traceback (most recent call last):\n'
                '  File: "./src/main.mys", line 5 in main\n'
                '    raise AnError("hi")\n'
                '\n'
                'AnError(message="hi")\n',
                'debug')

            with patch('sys.argv', ['mys', 'clean']):
                mys.cli.main()

            # Speed build, no tracebacks.
            with patch('sys.argv', ['mys', 'build']):
                mys.cli.main()

            self.run_app_assert('AnError(message="hi")\n',
                                'speed',
                                'Traceback (most recent call last):\n'
                                '  File: "./src/main.mys", line 5 in main\n'
                                '    raise AnError("hi")\n')



from ..utils import BuildConfig
from ..utils import add_coverage_argument
from ..utils import add_debug_symbols_argument
from ..utils import add_jobs_argument
from ..utils import add_no_ccache_argument
from ..utils import add_optimize_argument
from ..utils import add_unsafe_argument
from ..utils import add_url_argument
from ..utils import add_verbose_argument
from ..utils import build_app
from ..utils import build_prepare


def do_build(_parser, args, _mys_config):
    build_config = BuildConfig(args.debug,
                               args.verbose,
                               args.optimize,
                               args.debug_symbols,
                               args.no_ccache,
                               args.coverage,
                               args.unsafe,
                               args.jobs,
                               args.url)
    is_application, build_dir = build_prepare(build_config)
    build_app(build_config, is_application, build_dir)


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'build',
        description='Build the appliaction.')
    add_verbose_argument(subparser)
    add_jobs_argument(subparser)
    add_optimize_argument(subparser, 'speed')
    add_debug_symbols_argument(subparser)
    add_no_ccache_argument(subparser)
    add_url_argument(subparser)
    add_coverage_argument(subparser)
    add_unsafe_argument(subparser)
    subparser.set_defaults(func=do_build)



import glob
import os
import shutil
import sys
import tarfile
from tempfile import TemporaryDirectory

from ..utils import ERROR
from ..utils import BuildConfig
from ..utils import Spinner
from ..utils import add_jobs_argument
from ..utils import add_no_ccache_argument
from ..utils import add_url_argument
from ..utils import add_verbose_argument
from ..utils import box_print
from ..utils import build_app
from ..utils import build_prepare
from ..utils import read_package_configuration
from ..utils import run


def install_clean():
    if not os.path.exists('package.toml'):
        raise Exception('not a package')

    with Spinner(text='Cleaning'):
        shutil.rmtree('build', ignore_errors=True)

def install_download(build_config, package):
    command = [
        sys.executable, '-m', 'pip', 'download', f'mys-{package}'
    ]
    run(command, 'Downloading package', build_config.verbose)


def install_extract():
    archive = glob.glob('mys-*.tar.gz')[0]

    with Spinner(text='Extracting package'):
        with tarfile.open(archive) as fin:
            fin.extractall()

    os.remove(archive)


def install_build(build_config):
    config = read_package_configuration()
    is_application, build_dir = build_prepare(build_config, config)

    if not is_application:
        box_print(['There is no application to build in this package (src/main.mys ',
                   'missing).'],
                  ERROR)

        raise Exception()

    build_app(build_config, is_application, build_dir)

    return config

def install_install(root, _args, config):
    bin_dir = os.path.join(root, 'bin')
    bin_name = config['package']['name']
    src_file = 'build/speed-unsafe/app'
    dst_file = os.path.join(bin_dir, bin_name)

    with Spinner(text=f"Installing {bin_name} in {bin_dir}"):
        os.makedirs(bin_dir, exist_ok=True)
        shutil.copyfile(src_file, dst_file)
        shutil.copymode(src_file, dst_file)


def install_from_current_dirctory(build_config, root):
    install_clean()
    config = install_build(build_config)
    install_install(root, build_config, config)


def install_from_registry(build_config, package, root):
    with TemporaryDirectory()as tmp_dir:
        os.chdir(tmp_dir)
        install_download(build_config, package)
        install_extract()
        os.chdir(glob.glob('mys-*')[0])
        config = install_build(build_config)
        install_install(root, build_config, config)


def do_install(_parser, args, _mys_config):
    build_config = BuildConfig(args.debug,
                               args.verbose,
                               'speed',
                               False,
                               args.no_ccache,
                               False,
                               True,
                               args.jobs,
                               args.url)
    root = os.path.abspath(os.path.expanduser(args.root))

    if args.package is None:
        install_from_current_dirctory(build_config, root)
    else:
        install_from_registry(build_config, args.package, root)


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'install',
        description='Install an application from local package or registry.')
    add_verbose_argument(subparser)
    add_jobs_argument(subparser)
    add_no_ccache_argument(subparser)
    add_url_argument(subparser)
    subparser.add_argument('--root',
                           default='~/.local',
                           help='Root folder to install into (default: %(default)s.')
    subparser.add_argument(
        'package',
        nargs='?',
        help=('Package to install application from. Installs current package if '
              'not given.'))
    subparser.set_defaults(func=do_install)



from ..parser import ast
from .utils import CompileError
from .utils import indent
from .utils import make_shared_dict
from .utils import make_shared_list
from .utils import make_shared_set
from .utils import mys_to_cpp_type
from .utils import mys_to_cpp_type_param
from .utils import split_dict_mys_type


class Comprehension:

    def __init__(self, node, mys_type, visitor):
        self.node = node
        self.mys_type = mys_type
        self.visitor = visitor
        self.result_variable = None

    def value(self):
        raise NotImplementedError()

    def body(self):
        raise NotImplementedError()

    def generate(self):
        if len(self.node.generators) != 1:
            raise CompileError("only one for-loop allowed", self.node)

        self.visitor.in_comprehension = True
        context = self.visitor.context
        local_variables = []
        self_variable = None
        local_variables = [
            (k, v)
            for k, v in context.local_variables.items()
            if k != 'self'
        ]
        local_variables.sort()
        result_cpp_type = mys_to_cpp_type(self.mys_type, context)
        self.result_variable = context.unique('result')
        context.push()
        context.define_local_variable(self.result_variable,
                                      self.mys_type,
                                      self.node)

        if self_variable is not None:
            context.define_local_variable(self_variable[0],
                                          self_variable[1],
                                          self.node)
        generator = self.node.generators[0]

        if len(generator.ifs) > 1:
            raise CompileError("at most one if allowed", self.node)

        body = self.body()

        if generator.ifs:
            body = ast.If(test=generator.ifs[0],
                          body=[body],
                          orelse=[])

        code = '\n'.join([
            f'{result_cpp_type} {self.result_variable} = {self.value()};',
            self.visitor.visit_For(
                ast.fix_missing_locations(
                    ast.For(target=generator.target,
                            iter=generator.iter,
                            body=[body],
                            orelse=[]))),
            f'\nreturn {self.result_variable};'
        ])
        function_name = self.visitor.unique('list_comprehension')
        parameters = ', '.join([
            f'{mys_to_cpp_type_param(mys_type, context)} {name}'
            for name, mys_type in local_variables
        ])

        if context.class_name is None:
            code = '\n'.join([
                f'static {result_cpp_type} {function_name}({parameters})',
                '{',
                indent(code),
                '}'
            ])
        else:
            code = '\n'.join([
                f'{result_cpp_type} {context.class_name}::{function_name}('
                f'{parameters})',
                '{',
                indent(code),
                '}'
            ])
            context.method_comprehensions[context.class_name].append(
                f'{result_cpp_type} {function_name}({parameters});')

        context.comprehensions.append(code)
        context.pop()
        context.mys_type = self.mys_type
        parameters = ', '.join([name for name, _ in local_variables])
        self.visitor.in_comprehension = False

        return f'{function_name}({parameters})'


class ListComprehension(Comprehension):
    """Code generator for a list comprehension.

    """

    def value(self):
        result_item_cpp_type = self.visitor.mys_to_cpp_type(self.mys_type[0])

        return make_shared_list(result_item_cpp_type, '')

    def body(self):
        return ast.Expr(
            value=ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id=self.result_variable),
                    attr='append'),
                args=[self.node.elt]))

class DictComprehension(Comprehension):
    """Code generator for a dict comprehension.

    """

    def value(self):
        key_mys_type, value_mys_type = split_dict_mys_type(self.mys_type)
        key_cpp_type = self.visitor.mys_to_cpp_type(key_mys_type)
        value_cpp_type = self.visitor.mys_to_cpp_type(value_mys_type)

        return make_shared_dict(key_cpp_type, value_cpp_type, '')

    def body(self):
        return ast.Assign(
            targets=[
                ast.Subscript(
                    slice=self.node.key,
                    value=ast.Name(id=self.result_variable))
            ],
            value=self.node.value)

class SetComprehension(Comprehension):
    """Code generator for a set comprehension.

    """

    def value(self):
        result_item_cpp_type = self.visitor.mys_to_cpp_type(list(self.mys_type)[0])

        return make_shared_set(result_item_cpp_type, '')

    def body(self):
        return ast.Expr(
            value=ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id=self.result_variable),
                    attr='add'),
                args=[self.node.elt]))



import os

from ..utils import BuildConfig
from ..utils import add_coverage_argument
from ..utils import add_jobs_argument
from ..utils import add_no_ccache_argument
from ..utils import add_optimize_argument
from ..utils import add_unsafe_argument
from ..utils import add_url_argument
from ..utils import add_verbose_argument
from ..utils import build_prepare
from ..utils import create_coverage_report
from ..utils import run


def do_test(_parser, args, _mys_config):
    build_config = BuildConfig(args.debug,
                               args.verbose,
                               args.optimize,
                               True,
                               args.no_ccache,
                               args.coverage,
                               args.unsafe,
                               args.jobs,
                               args.url)
    _, build_dir = build_prepare(build_config)

    command = [
        'make', '-f', f'{build_dir}/Makefile', 'test', 'TEST=yes'
    ]

    if os.getenv('MAKEFLAGS') is None:
        command += ['-j', str(args.jobs)]

    if args.debug:
        command += ['TRANSPILE_DEBUG=--debug']

    if args.coverage:
        command += ['COVERAGE=yes']

    if args.unsafe:
        command += ['UNSAFE=yes']

    if args.optimize == 'debug':
        command += ['TRACEBACK=yes']

    if args.test_pattern is None:
        test_pattern = []
    else:
        test_pattern = [args.test_pattern]

    run(command, 'Building tests', args.verbose)
    run([f'./{build_dir}/test'] + test_pattern, 'Running tests', args.verbose)

    if args.coverage:
        create_coverage_report(['./src/**'])


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'test',
        description='Build and run tests.')
    add_verbose_argument(subparser)
    add_jobs_argument(subparser)
    add_optimize_argument(subparser, 'debug')
    add_no_ccache_argument(subparser)
    add_url_argument(subparser)
    add_coverage_argument(subparser)
    add_unsafe_argument(subparser)
    subparser.add_argument(
        'test_pattern',
        nargs='?',
        help=("Only run tests matching given pattern. '^' matches the "
              "beginning and '$' matches the end of the test name."))
    subparser.set_defaults(func=do_test)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Better tokenizing for coverage.py."""

import keyword
import re
import token
import tokenize


def phys_tokens(toks):
    """Return all physical tokens, even line continuations.

    tokenize.generate_tokens() doesn't return a token for the backslash that
    continues lines.  This wrapper provides those tokens so that we can
    re-create a faithful representation of the original source.

    Returns the same values as generate_tokens()

    """
    last_line = None
    last_lineno = -1
    last_ttext = None
    for ttype, ttext, (slineno, scol), (elineno, ecol), ltext in toks:
        if last_lineno != elineno:
            if last_line and last_line.endswith("\\\n"):
                # We are at the beginning of a new line, and the last line
                # ended with a backslash.  We probably have to inject a
                # backslash token into the stream. Unfortunately, there's more
                # to figure out.  This code::
                #
                #   usage = """\
                #   HEY THERE
                #   """
                #
                # triggers this condition, but the token text is::
                #
                #   '"""\\\nHEY THERE\n"""'
                #
                # so we need to figure out if the backslash is already in the
                # string token or not.
                inject_backslash = True
                if last_ttext.endswith("\\"):
                    inject_backslash = False
                elif ttype == token.STRING:
                    if "\n" in ttext and ttext.split('\n', 1)[0][-1] == '\\':
                        # It's a multi-line string and the first line ends with
                        # a backslash, so we don't need to inject another.
                        inject_backslash = False
                if inject_backslash:
                    # Figure out what column the backslash is in.
                    ccol = len(last_line.split("\n")[-2]) - 1
                    # Yield the token, with a fake token type.
                    yield (
                        99999, "\\\n",
                        (slineno, ccol), (slineno, ccol+2),
                        last_line
                        )
            last_line = ltext
        if ttype not in (tokenize.NEWLINE, tokenize.NL):
            last_ttext = ttext
        yield ttype, ttext, (slineno, scol), (elineno, ecol), ltext
        last_lineno = elineno


def source_token_lines(source):
    """Generate a series of lines, one for each line in `source`.

    Each line is a list of pairs, each pair is a token::

        [('key', 'def'), ('ws', ' '), ('nam', 'hello'), ('op', '('), ... ]

    Each pair has a token class, and the token text.

    If you concatenate all the token texts, and then join them with newlines,
    you should have your original `source` back, with two differences:
    trailing whitespace is not preserved, and a final line with no newline
    is indistinguishable from a final line with a newline.

    """

    ws_tokens = {token.INDENT, token.DEDENT, token.NEWLINE, tokenize.NL}
    line = []
    col = 0

    source = source.expandtabs(8).replace('\r\n', '\n')
    tokgen = generate_tokens(source)

    for ttype, ttext, (_, scol), (_, ecol), _ in phys_tokens(tokgen):
        mark_start = True
        for part in re.split('(\n)', ttext):
            if part == '\n':
                yield line
                line = []
                col = 0
                mark_end = False
            elif part == '':
                mark_end = False
            elif ttype in ws_tokens:
                mark_end = False
            else:
                if mark_start and scol > col:
                    line.append(("ws", u" " * (scol - col)))
                    mark_start = False
                tok_class = tokenize.tok_name.get(ttype, 'xx').lower()[:3]
                if ttype == token.NAME and keyword.iskeyword(ttext):
                    tok_class = "key"
                line.append((tok_class, part))
                mark_end = True
            scol = 0
        if mark_end:
            col = ecol

    if line:
        yield line


class CachedTokenizer:
    """A one-element cache around tokenize.generate_tokens.

    When reporting, coverage.py tokenizes files twice, once to find the
    structure of the file, and once to syntax-color it.  Tokenizing is
    expensive, and easily cached.

    This is a one-element cache so that our twice-in-a-row tokenizing doesn't
    actually tokenize twice.

    """
    def __init__(self):
        self.last_text = None
        self.last_tokens = None

    def generate_tokens(self, text):
        """A stand-in for `tokenize.generate_tokens`."""
        if text != self.last_text:
            self.last_text = text
            readline = iter(text.splitlines(True)).__next__
            self.last_tokens = list(tokenize.generate_tokens(readline))
        return self.last_tokens

# Create our generate_tokens cache as a callable replacement function.
generate_tokens = CachedTokenizer().generate_tokens


COOKIE_RE = re.compile(r"^[ \t]*#.*coding[:=][ \t]*([-\w.]+)", flags=re.MULTILINE)


def source_encoding(source):
    """Determine the encoding for `source`, according to PEP 263.

    `source` is a byte string: the text of the program.

    Returns a string, the name of the encoding.

    """
    readline = iter(source.splitlines(True)).__next__
    return tokenize.detect_encoding(readline)[0]



import difflib
import os
import re
import shutil
import subprocess
import sys
import unittest
from unittest.mock import patch

import mys.cli
from mys.transpiler import Source
from mys.transpiler import TranspilerError
from mys.transpiler import transpile


class TestCase(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        os.makedirs('tests/build', exist_ok=True)

    def assert_in(self, needle, haystack):
        try:
            self.assertIn(needle, haystack)
        except AssertionError:
            differ = difflib.Differ()
            diff = differ.compare(needle.splitlines(), haystack.splitlines())

            raise AssertionError(
                '\n' + '\n'.join([diffline.rstrip('\n') for diffline in diff]))

    def assert_not_in(self, needle, haystack):
        self.assertNotIn(needle, haystack)

    def assert_exception_string(self, cm, expected):
        self.assertEqual(expected, remove_ansi(str(cm.exception)))

    def assert_transpile_raises(self, source, error):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source(source)

        self.assert_exception_string(cm, error)

    def assert_file_exists(self, path):
        self.assertTrue(os.path.exists(path))

    def assert_file_not_exists(self, path):
        self.assertFalse(os.path.exists(path))

    def assert_files_equal(self, actual, expected):
        # os.makedirs(os.path.dirname(expected), exist_ok=True)
        # open(expected, 'w').write(open(actual, 'r').read())
        self.assertEqual(read_file(actual), read_file(expected))

    def assert_in_file(self, needle, haystack):
        self.assert_in(needle, read_file(haystack))

    def assert_not_in_file(self, needle, haystack):
        self.assertNotIn(needle, read_file(haystack))


def read_file(filename):
    with open(filename, 'r') as fin:
        return fin.read()


def remove_directory(name):
    if os.path.exists(name):
        shutil.rmtree(name)


def remove_ansi(string):
    ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')

    return ansi_escape.sub('', string)


def remove_build_directory(name):
    remove_directory('tests/build/' + name)


def create_new_package(package_name):
    with Path('tests/build'):
        remove_directory(package_name)

        command = [
            'mys', '-d', 'new',
            '--author', 'Test Er <test.er@mys.com>',
            package_name
        ]

        with patch('sys.argv', command):
            mys.cli.main()


def create_new_package_with_files(package_name, module_name, src_module_name=None):
    if src_module_name is None:
        src_module_name = module_name

    create_new_package(package_name)

    with Path('tests/build'):
        os.remove(f'{package_name}/src/lib.mys')
        os.remove(f'{package_name}/src/main.mys')
        shutil.copyfile(f'../../tests/files/{module_name}.mys',
                        f'{package_name}/src/{src_module_name}.mys')


def test_package(package_name, flags):
    if flags is None:
        flags = []

    with Path('tests/build/' + package_name):
        with patch('sys.argv', ['mys', '--debug', 'test', '--verbose'] + flags):
            mys.cli.main()


def build_and_test_module(module_name, flags=None):
    package_name = f'test_{module_name}'
    create_new_package_with_files(package_name, module_name)
    test_package(package_name, flags)


class Path:

    def __init__(self, new_dir):
        self.new_dir = new_dir
        self.old_dir = None

    def __enter__(self):
        self.old_dir = os.getcwd()
        os.chdir(self.new_dir)

        return self

    def __exit__(self, *args, **kwargs):
        os.chdir(self.old_dir)

        return False


def run_mys_command(command, path):
    env = os.environ
    env['PYTHONPATH'] = path
    proc = subprocess.run([sys.executable, '-m', 'mys'] + command,
                          capture_output=True,
                          text=True,
                          close_fds=False,
                          env=env)

    if proc.returncode != 0:
        print(proc.stdout)
        print(proc.stderr)

        raise Exception("Build error.")

    return proc.stdout, proc.stderr


def transpile_early_header(source, mys_path='', module_hpp=''):
    return transpile([Source(source,
                             mys_path=mys_path,
                             module_hpp=module_hpp)])[0][0]


def transpile_header(source, mys_path='', module_hpp=''):
    return transpile([Source(source,
                             mys_path=mys_path,
                             module_hpp=module_hpp)])[0][1]


def transpile_source(source,
                     mys_path='',
                     module='foo.lib',
                     module_hpp='foo/lib.mys.hpp',
                     has_main=False):
    return transpile([Source(source,
                             mys_path=mys_path,
                             module=module,
                             module_hpp=module_hpp,
                             has_main=has_main)])[0][2]



import os
import shutil

from ..utils import Spinner
from ..utils import read_package_configuration


def remove_file(path):
    try:
        os.remove(path)
    except FileNotFoundError:
        pass

def do_clean(_parser, _args, _mys_config):
    read_package_configuration()

    with Spinner(text='Cleaning'):
        shutil.rmtree('build', ignore_errors=True)
        shutil.rmtree('coverage', ignore_errors=True)
        remove_file('.coverage')
        remove_file('.mys-coverage.txt')


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'clean',
        description='Remove build output.')
    subparser.set_defaults(func=do_clean)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""File wrangling."""

import fnmatch
import hashlib
import ntpath
import os
import os.path
import re
import sys

from .misc import join_regex


def set_relative_directory():
    """Set the directory that `relative_filename` will be relative to."""
    global RELATIVE_DIR, CANONICAL_FILENAME_CACHE

    # The absolute path to our current directory.
    RELATIVE_DIR = os.path.normcase(abs_file(os.curdir) + os.sep)

    # Cache of results of calling the canonical_filename() method, to
    # avoid duplicating work.
    CANONICAL_FILENAME_CACHE = {}


def relative_filename(filename):
    """Return the relative form of `filename`.

    The file name will be relative to the current directory when the
    `set_relative_directory` was called.

    """
    fnorm = os.path.normcase(filename)
    if fnorm.startswith(RELATIVE_DIR):
        filename = filename[len(RELATIVE_DIR):]
    return unicode_filename(filename)


def canonical_filename(filename):
    """Return a canonical file name for `filename`.

    An absolute path with no redundant components and normalized case.

    """
    if filename not in CANONICAL_FILENAME_CACHE:
        cf = filename
        if not os.path.isabs(filename):
            for path in [os.curdir] + sys.path:
                if path is None:
                    continue
                ff = os.path.join(path, filename)
                try:
                    exists = os.path.exists(ff)
                except UnicodeError:
                    exists = False
                if exists:
                    cf = ff
                    break
        cf = abs_file(cf)
        CANONICAL_FILENAME_CACHE[filename] = cf
    return CANONICAL_FILENAME_CACHE[filename]


MAX_FLAT = 200

def flat_rootname(filename):
    """A base for a flat file name to correspond to this file.

    Useful for writing files about the code where you want all the files in
    the same directory, but need to differentiate same-named files from
    different directories.

    For example, the file a/b/c.py will return 'a_b_c_py'

    """
    name = ntpath.splitdrive(filename)[1]
    name = re.sub(r"[\\/.:]", "_", name)
    if len(name) > MAX_FLAT:
        hh = hashlib.sha1(name.encode('UTF-8')).hexdigest()
        name = name[-(MAX_FLAT-len(hh)-1):] + '_' + hh
    return name


def actual_path(filename):
    """The actual path for non-Windows platforms."""
    return filename


def unicode_filename(filename):
    """Return a Unicode version of `filename`."""
    return filename


def abs_file(path):
    """Return the absolute normalized form of `path`."""
    try:
        path = os.path.realpath(path)
    except UnicodeError:
        pass
    path = os.path.abspath(path)
    path = actual_path(path)
    path = unicode_filename(path)
    return path


RELATIVE_DIR = None
CANONICAL_FILENAME_CACHE = None
set_relative_directory()


def prep_patterns(patterns):
    """Prepare the file patterns for use in a `FnmatchMatcher`.

    If a pattern starts with a wildcard, it is used as a pattern
    as-is.  If it does not start with a wildcard, then it is made
    absolute with the current directory.

    If `patterns` is None, an empty list is returned.

    """
    prepped = []
    for pa in patterns or []:
        if pa.startswith(("*", "?")):
            prepped.append(pa)
        else:
            prepped.append(abs_file(pa))
    return prepped


class FnmatchMatcher:
    """A matcher for files by file name pattern."""
    def __init__(self, pats):
        self.pats = list(pats)
        self.re = fnmatches_to_regex(self.pats, case_insensitive=False)

    def __repr__(self):
        return "<FnmatchMatcher %r>" % self.pats

    def info(self):
        """A list of strings for displaying when dumping state."""
        return self.pats

    def match(self, fpath):
        """Does `fpath` match one of our file name patterns?"""
        return self.re.match(fpath) is not None


def fnmatches_to_regex(patterns, case_insensitive=False, partial=False):
    """Convert fnmatch patterns to a compiled regex that matches any of them.

    Slashes are always converted to match either slash or backslash, for
    Windows support, even when running elsewhere.

    If `partial` is true, then the pattern will match if the target string
    starts with the pattern. Otherwise, it must match the entire string.

    Returns: a compiled regex object.  Use the .match method to compare target
    strings.

    """
    regexes = (fnmatch.translate(pattern) for pattern in patterns)
    # Python3.7 fnmatch translates "/" as "/". Before that, it translates as "\/",
    # so we have to deal with maybe a backslash.
    regexes = (re.sub(r"\\?/", r"[\\\\/]", regex) for regex in regexes)

    if partial:
        # fnmatch always adds a \Z to match the whole string, which we don't
        # want, so we remove the \Z.  While removing it, we only replace \Z if
        # followed by paren (introducing flags), or at end, to keep from
        # destroying a literal \Z in the pattern.
        regexes = (re.sub(r'\\Z(\(\?|$)', r'\1', regex) for regex in regexes)

    flags = 0
    if case_insensitive:
        flags |= re.IGNORECASE
    compiled = re.compile(join_regex(regexes), flags=flags)

    return compiled



from mys.transpiler import TranspilerError

from .utils import TestCase
from .utils import build_and_test_module
from .utils import transpile_source


class Test(TestCase):

    def test_classes(self):
        build_and_test_module('classes')

    def test_class_in_function_should_fail(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source('def main():\n'
                             '    class A:\n'
                             '        pass\n',
                             mys_path='<unknown>',
                             has_main=True)

        self.assert_exception_string(
            cm,
            '  File "<unknown>", line 2\n'
            '        class A:\n'
            '        ^\n'
            'CompileError: class definitions are only allowed on module level\n')

    def test_undefined_class(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    Bar()\n',
            '  File "", line 2\n'
            '        Bar()\n'
            '        ^\n'
            "CompileError: undefined class/trait/enum 'Bar'\n")

    def test_test_decorator_only_allowed_on_functions(self):
        self.assert_transpile_raises(
            'class Class1:\n'
            '    @test\n'
            '    def foo(self):\n'
            '        pass\n',
            '  File "", line 2\n'
            '        @test\n'
            '         ^\n'
            "CompileError: invalid decorator 'test'\n")

    def test_non_snake_case_class_member(self):
        self.assert_transpile_raises(
            'class A:\n'
            '    Aa: i32',
            '  File "", line 2\n'
            '        Aa: i32\n'
            '        ^\n'
            "CompileError: class member names must be snake case\n")

    def test_non_pascal_case_class(self):
        self.assert_transpile_raises(
            'class apa:\n'
            '    pass\n',
            '  File "", line 1\n'
            '    class apa:\n'
            '    ^\n'
            "CompileError: class names must be pascal case\n")

    def test_invalid_decorator(self):
        self.assert_transpile_raises(
            '@foobar\n'
            'class Foo:\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @foobar\n'
            '     ^\n'
            "CompileError: invalid decorator 'foobar'\n")

    def test_wrong_method_parameter_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self, a: string):\n'
            '        pass\n'
            'def bar():\n'
            '    Foo().foo(True)\n',
            '  File "", line 5\n'
            '        Foo().foo(True)\n'
            '                  ^\n'
            "CompileError: expected a 'string', got a 'bool'\n")

    def test_call_member_method(self):
        source = transpile_source('class Foo:\n'
                                  '    def fam(self):\n'
                                  '        pass\n'
                                  'class Bar:\n'
                                  '    foo: Foo\n'
                                  'def foo2(bar: Bar):\n'
                                  '    bar.foo.fam()')

        self.assert_in('bar->foo->fam();', source)

    def test_assign_to_self_1(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self):\n'
            '        self = Foo()\n',
            '  File "", line 3\n'
            '            self = Foo()\n'
            '            ^\n'
            "CompileError: it's not allowed to assign to 'self'\n")

    def test_assign_to_self_2(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self):\n'
            '        self: u8 = 1\n',
            '  File "", line 3\n'
            '            self: u8 = 1\n'
            '            ^\n'
            "CompileError: redefining variable 'self'\n")

    def test_class_functions_not_implemented(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo():\n'
            '        pass\n',
            '  File "", line 2\n'
            "        def foo():\n"
            '        ^\n'
            "CompileError: class functions are not yet implemented\n")

    def test_compare_wrong_types_19(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            'class Bar:\n'
            '    pass\n'
            'def foo():\n'
            '    if Foo() is Bar():\n'
            '        pass\n',
            '  File "", line 6\n'
            "        if Foo() is Bar():\n"
            '           ^\n'
            "CompileError: types 'foo.lib.Foo' and 'foo.lib.Bar' differs\n")

    def test_class_has_no_member_1(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    value: i32\n'
            'def foo(v: Foo):\n'
            '    print(v.missing)\n',
            '  File "", line 4\n'
            '        print(v.missing)\n'
            '              ^\n'
            "CompileError: class 'foo.lib.Foo' has no member 'missing'\n")

    def test_class_has_no_member_2(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    value: i32\n'
            '    def foo(self):\n'
            '        print(self.missing)\n',
            '  File "", line 4\n'
            '            print(self.missing)\n'
            '                  ^\n'
            "CompileError: class 'foo.lib.Foo' has no member 'missing'\n")

    def test_class_has_no_member_3(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: i32\n'
            'def foo():\n'
            '    print(Foo(1).b)\n',
            '  File "", line 4\n'
            '        print(Foo(1).b)\n'
            '              ^\n'
            "CompileError: class 'foo.lib.Foo' has no member 'b'\n")

    def test_class_has_no_member_4(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self):\n'
            '        if self.a == 1:\n'
            '            pass\n',
            '  File "", line 3\n'
            '            if self.a == 1:\n'
            '               ^\n'
            "CompileError: class 'foo.lib.Foo' has no member 'a'\n")

    def test_class_private_member(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    _a: i32\n'
            'def foo():\n'
            '    print(Foo()._a)\n',
            '  File "", line 4\n'
            '        print(Foo()._a)\n'
            '              ^\n'
            "CompileError: class 'foo.lib.Foo' member '_a' is private\n")

    def test_unknown_class_member_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: Bar\n',
            '  File "", line 2\n'
            '        a: Bar\n'
            '           ^\n'
            "CompileError: undefined type 'Bar'\n")

    def test_class_init_too_many_parameters(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: i32\n'
            'def foo():\n'
            '    print(Foo(1, 2))\n',
            '  File "", line 4\n'
            '        print(Foo(1, 2))\n'
            '              ^\n'
            "CompileError: expected 1 parameter, got 2\n")

    def test_class_init_too_few_parameters_with_default(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: i32\n'
            '    c: string\n'
            '    def __init__(self, a: i32, b: i32 = 5, c: string = None):\n'
            '        self.a = a + b\n'
            '        self.c = c\n'
            'def foo():\n'
            '    print(Foo())\n',
            '  File "", line 8\n'
            '        print(Foo())\n'
            '              ^\n'
            "CompileError: parameter 'a: i32' not given\n")

    def test_class_init_wrong_parameter_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: (bool, string)\n'
            'def foo():\n'
            '    print(Foo(("", 1)))\n',
            '  File "", line 4\n'
            '        print(Foo(("", 1)))\n'
            '                   ^\n'
            "CompileError: expected a 'bool', got a 'string'\n")

    def test_assign_to_class_call(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('class Foo:\n'
                             '    def bar() -> i32:\n'
                             '        return 1\n'
                             'def foo():\n'
                             '    Foo() = 1\n')

        self.assert_exception_string(
            cm,
            '  File "<string>", line 5\n'
            '    Foo() = 1\n'
            '    ^\n'
            "SyntaxError: cannot assign to function call\n")

    def test_assign_to_method_call(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('class Foo:\n'
                             '    def bar() -> i32:\n'
                             '        return 1\n'
                             'def foo():\n'
                             '    Foo().bar() = 1\n')

        self.assert_exception_string(
            cm,
            '  File "<string>", line 5\n'
            '    Foo().bar() = 1\n'
            '    ^\n'
            "SyntaxError: cannot assign to function call\n")

    def test_class_member_default(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    a: i32 = 1\n',
            '  File "", line 2\n'
            '        a: i32 = 1\n'
            '                 ^\n'
            "CompileError: class members cannot have default values\n")

    def test_wrong_class_method_parameter_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self, v: bool):\n'
            '        pass\n'
            'def foo(v: Foo):\n'
            '    v.foo("")\n',
            '  File "", line 5\n'
            '        v.foo("")\n'
            '              ^\n'
            "CompileError: expected a 'bool', got a 'string'\n")

    def test_positive_class(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            'def foo():\n'
            '    print(+Foo())\n',
            '  File "", line 4\n'
            '        print(+Foo())\n'
            '              ^\n'
            "CompileError: unary '+' can only operate on numbers\n")

    def test_method_call_in_assert(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def get_self(self) -> Foo:\n'
            '        return self\n'
            '    def get_same(self, this: Foo) -> Foo:\n'
            '        return this\n'
            'def foo():\n'
            '    x = Foo()\n'
            '    assert x is x.get_self().same(x)\n',
            '  File "", line 8\n'
            '        assert x is x.get_self().same(x)\n'
            '                    ^\n'
            "CompileError: class 'foo.lib.Foo' has no method 'same'\n")

    def test_import_after_class_definition(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            'from bar import fie\n',
            '  File "", line 3\n'
            '    from bar import fie\n'
            '    ^\n'
            "CompileError: imports must be at the beginning of the file\n")

    def test_class_type_mismatch(self):
        self.assert_transpile_raises(
            'class Add:\n'
            '    def calc(self, a: (One, u8)) -> u8:\n'
            '        return 2 * a[0].calc()\n'
            'class One:\n'
            '    def calc(self) -> u8:\n'
            '        return 1\n'
            'def foo():\n'
            '    x = Add()\n'
            '    assert x.calc(One()) == 2\n',
            '  File "", line 9\n'
            '        assert x.calc(One()) == 2\n'
            '                      ^\n'
            "CompileError: expected a '(foo.lib.One, u8)', got a 'foo.lib.One'\n")

    def test_operator_overloading_wrong_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def __add__(self, other: f64) -> f64:\n'
            '        return other\n'
            'def foo():\n'
            '    a = Foo()\n'
            '    print(a + 1)\n',
            '  File "", line 6\n'
            '        print(a + 1)\n'
            '                  ^\n'
            "CompileError: cannot convert integer to 'f64'\n")

    def test_operator_overloading_no_parameter(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def __add__(self) -> f64:\n'
            '        return 1.0\n',
            '  File "", line 2\n'
            '        def __add__(self) -> f64:\n'
            '        ^\n'
            "CompileError: __add__ must take exactly one parameter\n")

    def test_operator_overloading_no_return_value(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def __add__(self, other: bool):\n'
            '        pass\n',
            '  File "", line 2\n'
            '        def __add__(self, other: bool):\n'
            '        ^\n'
            "CompileError: __add__ must return a value\n")

    def test_operator_not_overloaded(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            'def foo():\n'
            '    print(Foo() + 1)\n',
            '  File "", line 4\n'
            '        print(Foo() + 1)\n'
            '              ^\n'
            "CompileError: class 'foo.lib.Foo' has no method '__add__'\n")

    def test_invalid_member_name(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    self.apa: i64\n',
            '  File "", line 2\n'
            '        self.apa: i64\n'
            '        ^\n'
            "CompileError: invalid syntax\n")

    def test_method_does_not_return(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self) -> i64:\n'
            '        pass\n',
            '  File "", line 2\n'
            '        def foo(self) -> i64:\n'
            '        ^\n'
            "CompileError: missing return or raise\n")



from collections import Counter
from textwrap import indent

from ....parser import ast
from ....transpiler.definitions import TypeVisitor
from ....transpiler.utils import OPERATORS
from ....transpiler.utils import format_mys_type
from ....transpiler.utils import has_docstring
from .comments_reader import CommentsReader
from .utils import get_function_or_class_node_start
from .utils import get_source


class SourceStyler(ast.NodeVisitor):

    def __init__(self):
        self.source_lines = None
        self.tree = None
        self.comments_reader = None
        self.local_imports = None
        self.other_imports = None
        self.code = None

    def style(self, source_lines, tree, comments):
        """Returns the styled source code from given source code, AST and
        extracted comments and blank lines.

        """

        self.source_lines = source_lines
        self.tree = tree
        self.comments_reader = CommentsReader(comments)
        self.local_imports = []
        self.other_imports = []
        self.code = []

        for node in self.tree.body:
            if isinstance(node, ast.ImportFrom):
                self._style_import_from(node)
            elif isinstance(node, ast.FunctionDef):
                self._style_function(node)
            elif isinstance(node, ast.ClassDef):
                self._style_class(node)
            elif isinstance(node, ast.AnnAssign):
                self._style_ann_assign(node)
            else:
                self._style_other(node)

        imports = []

        if self.other_imports:
            imports += sorted(self.other_imports)

        if self.local_imports:
            imports += sorted(self.local_imports)

        if imports:
            imports.append('')

        self.code = imports + self.code
        self.code += self.comments_reader.get_remaining(self.get_prev_line())

        if len(self.code) == 0 or self.code[-1] != '':
            self.code.append('')

        return '\n'.join([line.rstrip() for line in self.code])

    def get_prev_line(self):
        if len(self.code) == 0:
            return ''
        else:
            return self.code[-1]

    def get_comments_before(self, lineno):
        prev_line = self.get_prev_line()
        comments = self.comments_reader.get_before(lineno, prev_line)

        return comments

    def get_inline_comment(self, lineno):
        comment = self.comments_reader.get_at(lineno)

        if comment:
            comment = '  ' + comment.strip()

        return comment

    def ensure_blank_line(self):
        if len(self.code) > 0:
            if self.get_prev_line() != '':
                self.code.append('')

    def _style_import_from(self, node):
        if node.module is None:
            module = ''
        else:
            module = node.module

        module = '.' * node.level + module
        names = []

        for item in node.names:
            if item.asname is None:
                name = item.name
            else:
                name = f'{item.name} as {item.asname}'

            names.append(name)

        comment = self.get_inline_comment(node.end_lineno)
        names = ', '.join(names)
        styled_code = f'from {module} import {names}{comment}'

        if module[0] == '.':
            self.local_imports.append(styled_code)
        else:
            self.other_imports.append(styled_code)

    def _style_function(self, node):
        lineno = get_function_or_class_node_start(node)[0]
        comments = self.get_comments_before(lineno)

        if comments:
            self.code += comments
        else:
            self.ensure_blank_line()

        self.code += self.visit_decorator_list(node.decorator_list)
        self.code += get_source(self.source_lines,
                                node.lineno,
                                node.col_offset,
                                node.end_lineno,
                                node.end_col_offset)[1]
        self.code[-1] += self.get_inline_comment(node.end_lineno)

    def _style_class(self, node):
        lineno = get_function_or_class_node_start(node)[0]
        comments = self.get_comments_before(lineno)

        if comments:
            self.code += comments
        else:
            self.ensure_blank_line()

        members = []
        methods = []

        for i, item in enumerate(node.body):
            comments = self.get_comments_before(item.lineno)

            if isinstance(item, ast.FunctionDef):
                methods.append('')
                methods += comments
                methods += self.visit_decorator_list(item.decorator_list, '    ')
                methods += get_source(self.source_lines,
                                      item.lineno,
                                      item.col_offset - 4,
                                      item.end_lineno,
                                      item.end_col_offset)[1]
                methods[-1] += self.get_inline_comment(item.end_lineno)
            else:
                members += comments

                if isinstance(item, ast.AnnAssign):
                    mys_type = format_mys_type(TypeVisitor().visit(item.annotation))
                    members.append(f'    {item.target.id}: {mys_type}')
                else:
                    members += get_source(self.source_lines,
                                          item.lineno,
                                          item.col_offset - 4,
                                          item.end_lineno,
                                          item.end_col_offset)[1]

                members[-1] += self.get_inline_comment(item.end_lineno)

                if i == 0 and has_docstring(node):
                    members.append('')

        bases = ', '.join([base.id for base in node.bases])

        if bases:
            bases = f'({bases})'

        self.code += self.visit_decorator_list(node.decorator_list)
        self.code.append(f'class {node.name}{bases}:')
        self.code += members

        if self.get_prev_line() == '':
            self.code += methods[1:]
        else:
            self.code += methods

    def _style_ann_assign(self, node):
        self.code += self.get_comments_before(node.lineno)
        source = [self.visit(node.value)]
        mys_type = format_mys_type(TypeVisitor().visit(node.annotation))
        code = f'{node.target.id}: {mys_type} = '
        source[0] = code + source[0]
        self.code += source

    def _style_other(self, node):
        # Just get everything the node spans for now.
        self.code += self.get_comments_before(node.lineno)
        self.code += get_source(self.source_lines,
                                node.lineno,
                                node.col_offset,
                                node.end_lineno,
                                node.end_col_offset)[1]

    def visit_decorator_list(self, node, prefix=''):
        decorators = []

        for item in node:
            decorators.append(f'{prefix}@{self.visit(item)}')

        return decorators

    def visit_Name(self, node):
        return node.id

    def visit_Call(self, node):
        params = [self.visit(item) for item in node.args]
        params = ', '.join(params)

        return f'{node.func.id}({params})'

    def visit_list_table(self, node, items):
        if not items:
            return None

        number_of_items_per_row = Counter([item.lineno for item in node.elts])

        if len(number_of_items_per_row) <= 1:
            return None

        rows = sorted(number_of_items_per_row.items())
        leading_rows = rows[:-1]
        last_row = rows[-1]
        items_per_row = leading_rows[0][1]
        linenos = [
            row[0]
            for row in leading_rows
            if row[1] == items_per_row
        ]

        if len(linenos) != len(leading_rows):
            return None

        number_of_items_per_row = leading_rows[0][1]

        if last_row[1] > number_of_items_per_row:
            return None

        column_widths = [0] * number_of_items_per_row

        for i in range(len(items)):
            column_index = i % number_of_items_per_row
            width = len(items[i])

            if width > column_widths[column_index]:
                column_widths[column_index] = width

        column_formats = [f'{{:>{width}}}'for width in column_widths]
        code = []

        for i in range(len(rows)):
            pos = i * number_of_items_per_row
            row_items = []

            for j, item in enumerate(items[pos:pos + number_of_items_per_row]):
                row_items.append(column_formats[j].format(item))

            code.append(', '.join(row_items))

        code = indent(',\n'.join(code), '    ')
        code = f'\n{code}\n'

        return code

    def visit_List(self, node):
        all_items_single_line = True
        items = []

        for item in node.elts:
            code = self.visit(item)

            if '\n' in code:
                all_items_single_line = False

            items.append(code)

        if all_items_single_line:
            code = self.visit_list_table(node, items)

            if code is None:
                code = ', '.join(items)

                # Not taking account of the starting column.
                if len(code) > 70:
                    code = ',\n'.join([f'    {item}' for item in items])
                    code = f'\n{code}\n'
        else:
            items = [indent(item, '    ') for item in items]
            code = ',\n'.join(items)
            code = f'\n{code}\n'

        return f'[{code}]'

    def visit_Dict(self, node):
        items = []

        for key, value in zip(node.keys, node.values):
            key = self.visit(key)
            value = self.visit(value)

            if '\n' in value:
                value = indent(value, '    ').strip()

            items.append(f'    {key}: {value}')

        code = ',\n'.join(items)

        if code:
            code = f'\n{code}\n'

        return f'{{{code}}}'

    def visit_Tuple(self, node):
        all_items_single_line = True
        items = []

        for item in node.elts:
            code = self.visit(item)

            if '\n' in code:
                all_items_single_line = False

            items.append(code)

        if all_items_single_line:
            code = ', '.join(items)

            # Not taking account of the starting column.
            if len(code) > 70:
                code = ',\n'.join([f'    {item}' for item in items])
                code = f'\n{code}\n'
        else:
            items = [indent(item, '    ') for item in items]
            code = ',\n'.join(items)
            code = f'\n{code}\n'

        return f'({code})'

    def visit_Constant(self, node):
        return '\n'.join(get_source(self.source_lines,
                                    node.lineno,
                                    node.col_offset,
                                    node.end_lineno,
                                    node.end_col_offset)[1])

    def visit_UnaryOp(self, node):
        return '-' + self.visit(node.operand)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        op = OPERATORS[type(node.op)]
        right = self.visit(node.right)

        return f'{left} {op} {right}'

    def generic_visit(self, node):
        raise Exception('unsupported node type {type(node)} in the styler')



import shutil
import signal
from unittest.mock import patch

import pexpect

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import remove_build_directory


class Test(TestCase):

    def test_sigint(self):
        name = 'test_signals_sigint'
        remove_build_directory(name)

        shutil.copytree('tests/files/signals', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            try:
                with patch('sys.argv', ['mys', 'test', 'no-test']):
                    mys.cli.main()
            except SystemExit:
                pass

            test = pexpect.spawn('./build/debug/test test_sigint',
                                 encoding='utf-8',
                                 codec_errors='replace')
            test.expect('test_sigint started')
            test.kill(signal.SIGINT)
            test.expect('InterruptError()')
            test.wait()
            self.assertEqual(test.exitstatus, 0)



_CPP_RESERVED = set([
    'alignas',
    'alignof',
    'and',
    'and_eq',
    'asm',
    'atomic_cancel',
    'atomic_commit',
    'atomic_noexcept',
    'auto',
    'bitand',
    'bitor',
    'bool',
    'break',
    'case',
    'catch',
    'char',
    'char8_t',
    'char16_t',
    'char32_t',
    'class',
    'compl',
    'concept',
    'const',
    'consteval',
    'constexpr',
    'constinit',
    'const_cast',
    'continue',
    'co_await',
    'co_return',
    'co_yield',
    'decltype',
    'default',
    'delete',
    'do',
    'double',
    'dynamic_cast',
    'else',
    'enum',
    'explicit',
    'export',
    'extern',
    'false',
    'float',
    'for',
    'friend',
    'goto',
    'if',
    'inline',
    'int',
    'long',
    'mutable',
    'namespace',
    'new',
    'noexcept',
    'not',
    'not_eq',
    'nullptr',
    'operator',
    'or',
    'or_eq',
    'private',
    'protected',
    'public',
    'reflexpr',
    'register',
    'reinterpret_cast',
    'requires',
    'return',
    'short',
    'signed',
    'sizeof',
    'static',
    'static_assert',
    'static_cast',
    'struct',
    'switch',
    'synchronized',
    'template',
    'this',
    'thread_local',
    'throw',
    'true',
    'try',
    'typedef',
    'typeid',
    'typename',
    'union',
    'unsigned',
    'using',
    'virtual',
    'void',
    'volatile',
    'wchar_t',
    'while',
    'xor',
    'xor_eq',
])


def make_cpp_safe_name(name):
    """Returns given name with a prefix if it is a reserved C++ word.

    """

    if name in _CPP_RESERVED:
        name = f'__cpp_{name}'

    return name



import subprocess

from colors import cyan
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import PythonLexer

from ..utils import BULB
from ..utils import BuildConfig
from ..utils import add_coverage_argument
from ..utils import add_debug_symbols_argument
from ..utils import add_jobs_argument
from ..utils import add_no_ccache_argument
from ..utils import add_optimize_argument
from ..utils import add_unsafe_argument
from ..utils import add_url_argument
from ..utils import add_verbose_argument
from ..utils import box_print
from ..utils import build_app
from ..utils import build_prepare
from ..utils import create_coverage_report


def run_app(args, verbose, build_dir):
    command = f'./{build_dir}/app'

    if verbose:
        print(command)

    subprocess.run([command] + args, check=True)


def style_source(code):
    return highlight(code,
                     PythonLexer(),
                     Terminal256Formatter(style='monokai')).rstrip()


def do_run(_parser, args, _mys_config):
    build_config = BuildConfig(args.debug,
                               args.verbose,
                               args.optimize,
                               args.debug_symbols,
                               args.no_ccache,
                               args.coverage,
                               args.unsafe,
                               args.jobs,
                               args.url)
    is_application, build_dir = build_prepare(build_config)

    if is_application:
        build_app(build_config, True, build_dir)
        run_app(args.args, args.verbose, build_dir)

        if args.coverage:
            create_coverage_report()
    else:
        main_1 = style_source('def main():\n')
        main_2 = style_source("    print('Hello, world!')\n")
        func = style_source('main()')
        box_print([
            f"This package is not executable. Create '{cyan('src/main.mys')}' and",
            f"implement '{func}' to make the package executable.",
            '',
            main_1,
            main_2], BULB)

        raise Exception()


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'run',
        description='Build and run the application.')
    add_verbose_argument(subparser)
    add_jobs_argument(subparser)
    add_optimize_argument(subparser, 'speed')
    add_debug_symbols_argument(subparser)
    add_no_ccache_argument(subparser)
    add_url_argument(subparser)
    add_coverage_argument(subparser)
    add_unsafe_argument(subparser)
    subparser.add_argument('args', nargs='*')
    subparser.set_defaults(func=do_run)



from ..parser import ast
from .utils import INTEGER_TYPES
from .utils import CompileError
from .utils import is_public


def member_title(delim, member_name, pre=''):
    return ast.Constant(value=f'{delim}{member_name}={pre}')


def formatted_value_member(member_name):
    return ast.FormattedValue(
        value=ast.Attribute(
            value=ast.Name(id='self', ctx=ast.Load()),
            attr=member_name,
            ctx=ast.Load()),
        conversion=-1)


class ClassTransformer(ast.NodeTransformer):
    """Traverses the AST and adds missing __init__(), __del__() and
    __str__() methods to classes.

    """

    def default_member_value(self, annotation):
        """Default value of private member.

        """

        value = None

        if isinstance(annotation, ast.Name):
            type_name = annotation.id

            if type_name in INTEGER_TYPES:
                value = 0
            elif type_name in ['f32', 'f64']:
                value = 0.0
            elif type_name == 'bool':
                value = False
            elif type_name == 'char':
                value = ('', '', '')

        return ast.Constant(value=value)

    def add_init(self, node, members):
        """Add __init__() to the class as it was missing.

        """

        args = ast.arguments(posonlyargs=[],
                             args=[ast.arg(arg='self')],
                             vararg=None,
                             kwonlyargs=[],
                             kw_defaults=[],
                             kwarg=None,
                             defaults=[])
        body = []

        for member in members:
            member_name = member.target.id

            if is_public(member_name):
                args.args.append(ast.arg(arg=member_name,
                                         annotation=member.annotation))
                value = ast.Name(id=member_name)
            else:
                value = self.default_member_value(member.annotation)

            body.append(
                ast.Assign(
                    targets=[
                        ast.Attribute(
                            value=ast.Name(id='self', ctx=ast.Load()),
                            attr=member_name,
                            ctx=ast.Store())
                    ],
                    value=value))

        if not body:
            body.append(ast.Pass())

        node.body.append(ast.FunctionDef(name='__init__',
                                         args=args,
                                         body=body,
                                         decorator_list=[]))

    def add_del(self, node):
        """Add __del__() to the class as it was missing.

        """

        node.body.append(
            ast.FunctionDef(name='__del__',
                            args=ast.arguments(posonlyargs=[],
                                               args=[ast.arg(arg='self')],
                                               vararg=None,
                                               kwonlyargs=[],
                                               kw_defaults=[],
                                               kwarg=None,
                                               defaults=[]),
                            body=[],
                            decorator_list=[]))

    def add_str_string(self, body, values, delim, member_name):
        body += [
            ast.If(
                test=ast.Compare(
                    left=ast.Attribute(
                        value=ast.Name(id='self', ctx=ast.Load()),
                        attr=member_name,
                        ctx=ast.Load()),
                    ops=[ast.Is()],
                    comparators=[ast.Constant(value=None)]),
                body=[
                    ast.Assign(
                        targets=[ast.Name(id=member_name, ctx=ast.Store())],
                        value=ast.Constant(value='None'))
                ],
                orelse=[
                    ast.Assign(
                        targets=[ast.Name(id=member_name, ctx=ast.Store())],
                        value=ast.JoinedStr(
                            values=[
                                ast.Constant(value='"'),
                                ast.FormattedValue(
                                    value=ast.Attribute(
                                        value=ast.Name(id='self', ctx=ast.Load()),
                                        attr=member_name,
                                        ctx=ast.Load()),
                                    conversion=-1),
                                ast.Constant(value='"')
                            ]))
                ])
        ]
        values += [
            member_title(delim, member_name),
            ast.FormattedValue(
                value=ast.Name(id=member_name, ctx=ast.Load()),
                conversion=-1)
        ]

    def add_str_char(self, values, delim, member_name):
        values += [
            member_title(delim, member_name, "'"),
            formatted_value_member(member_name),
            ast.Constant(value="'")
        ]

    def add_str_other(self, values, delim, member_name):
        values += [
            member_title(delim, member_name),
            formatted_value_member(member_name)
        ]

    def add_str(self, node, members):
        """Add __str__() to the class as it was missing.

        """

        body = []
        values = [ast.Constant(value=f'{node.name}(')]
        delim = ''

        for member in members:
            member_name = member.target.id

            if isinstance(member.annotation, ast.Name):
                if member.annotation.id == 'string':
                    self.add_str_string(body, values, delim, member_name)
                elif member.annotation.id == 'char':
                    self.add_str_char(values, delim, member_name)
                else:
                    self.add_str_other(values, delim, member_name)
            else:
                self.add_str_other(values, delim, member_name)

            delim = ', '

        values.append(ast.Constant(value=')'))
        body.append(ast.Return(value=ast.JoinedStr(values=values)))
        node.body.append(
            ast.FunctionDef(name='__str__',
                            args=ast.arguments(posonlyargs=[],
                                               args=[ast.arg(arg='self')],
                                               vararg=None,
                                               kwonlyargs=[],
                                               kw_defaults=[],
                                               kwarg=None,
                                               defaults=[]),
                            body=body,
                            decorator_list=[],
                            returns=ast.Name(id='string', ctx=ast.Load())))

    def visit_ClassDef(self, node):
        init_found = False
        del_found = False
        str_found = False
        members = []

        # Traits and enums are not yet keywords and should not have
        # __init__ added.
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Name):
                if decorator.id in ['trait', 'enum']:
                    return node
            elif isinstance(decorator, ast.Call):
                if decorator.func.id == 'enum':
                    return node

        for item in node.body:
            if isinstance(item, ast.FunctionDef):
                if item.name == '__init__':
                    init_found = True
                elif item.name == '__del__':
                    del_found = True
                elif item.name == '__str__':
                    str_found = True
            elif isinstance(item, ast.AnnAssign):
                if not isinstance(item.target, ast.Name):
                    raise CompileError('invalid syntax', item)

                members.append(item)

        if not init_found:
            self.add_init(node, members)

        if not del_found:
            self.add_del(node)

        if not str_found:
            self.add_str(node, members)

        return node



import shutil
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import create_new_package
from .utils import read_file
from .utils import remove_build_directory


class Test(TestCase):

    def assert_files_equal(self, actual, expected):
        # os.makedirs(os.path.dirname(expected), exist_ok=True)
        # open(expected, 'w').write(open(actual, 'r').read())
        self.assertEqual(read_file(actual), read_file(expected))

    def test_foo_build_with_local_path_dependencies(self):
        package_name = 'test_foo_build_with_local_path_dependencies'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            # Add dependencies.
            with open('package.toml', 'a') as fout:
                fout.write('bar = { path = "../../files/bar" }\n'
                           'fie = { path = "../../files/fie" }\n')

            # Run.
            with patch('sys.argv', ['mys', '-d', 'run', '-v']):
                mys.cli.main()

            self.assert_file_exists(
                f'build/speed/cpp/include/{package_name}/main.mys.hpp')
            self.assert_file_exists(
                f'build/speed/cpp/src/{package_name}/main.mys.cpp')
            self.assert_file_exists('build/speed/cpp/include/bar/lib.mys.hpp')
            self.assert_file_exists('build/speed/cpp/src/bar/lib.mys.cpp')
            self.assert_file_exists('build/speed/cpp/include/fie/lib.mys.hpp')
            self.assert_file_exists('build/speed/cpp/src/fie/lib.mys.cpp')
            self.assert_file_exists('./build/speed/app')

    def test_foo_build_with_dependencies(self):
        # New.
        package_name = 'test_foo_build_with_dependencies'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            # Add dependencies.
            with open('package.toml', 'a') as fout:
                fout.write('bar = "0.3.0"\n')

            # Import from bar.
            with open('src/main.mys', 'w') as fout:
                print('from bar import hello', file=fout)
                print('', file=fout)
                print('def main():', file=fout)
                print('    v = "3.14"', file=fout)
                print('    hello(v)', file=fout)

            # Run.
            with patch('sys.argv', ['mys', '-d', 'run']):
                mys.cli.main()

            self.assert_file_exists(
                f'build/speed/cpp/include/{package_name}/main.mys.hpp')
            self.assert_file_exists(
                f'build/speed/cpp/src/{package_name}/main.mys.cpp')
            self.assert_file_exists('build/speed/cpp/include/bar/lib.mys.hpp')
            self.assert_file_exists('build/speed/cpp/src/bar/lib.mys.cpp')
            self.assert_file_exists('./build/speed/app')

    def test_c_dependencies(self):
        name = 'test_c_dependencies'
        remove_build_directory(name)
        shutil.copytree('tests/files/c_dependencies', f'tests/build/{name}')

        with Path(f'tests/build/{name}/mypkg'):
            with patch('sys.argv', ['mys', '-d', 'test']):
                mys.cli.main()

    def test_c_dependencies_not_found(self):
        name = 'test_c_dependencies_not_found'
        remove_build_directory(name)
        shutil.copytree('tests/files/c_dependencies_not_found', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            with patch('sys.argv', ['mys', 'test']):
                with self.assertRaises(SystemExit):
                    mys.cli.main()



from .utils import TestCase


class Test(TestCase):

    def test_value_if_cond_else_value_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(1 if 1 else 2)\n',
            '  File "", line 2\n'
            '        print(1 if 1 else 2)\n'
            '                   ^\n'
            "CompileError: expected a 'bool', got a 'i64'\n")

    def test_value_if_cond_else_value_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(1 if True else "")\n',
            '  File "", line 2\n'
            '        print(1 if True else "")\n'
            '                             ^\n'
            "CompileError: expected a 'i64', got a 'string'\n")

    def test_if_cond_as_non_bool(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if 1:\n'
            '        pass\n',
            '  File "", line 2\n'
            '        if 1:\n'
            '           ^\n'
            "CompileError: expected a 'bool', got a 'i64'\n")

    def test_bare_name_in_if(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: i32 = 0\n'
            '    if True:\n'
            '        a\n',
            '  File "", line 4\n'
            '            a\n'
            '            ^\n'
            "CompileError: bare name\n")

    def test_bare_name_in_else(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: i32 = 0\n'
            '    if True:\n'
            '        pass\n'
            '    else:\n'
            '        a\n',
            '  File "", line 6\n'
            '            a\n'
            '            ^\n'
            "CompileError: bare name\n")

    def test_variable_defined_in_if_can_not_be_used_after(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if True:\n'
            '        v = 1\n'
            '    print(v)\n',
            '  File "", line 4\n'
            '        print(v)\n'
            '              ^\n'
            "CompileError: undefined variable 'v'\n")

    def test_if_else_different_variable_type_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        x: i8 = 1\n'
            '    else:\n'
            '        x: i16 = 2\n'
            '    print(x)\n',
            '  File "", line 6\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_if_else_different_variable_type_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        x = 1\n'
            '    elif True:\n'
            '        x = 2\n'
            '    else:\n'
            '        if True:\n'
            '            x = ""\n'
            '        else:\n'
            '            x = 3\n'
            '    print(x)\n',
            '  File "", line 11\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_value_in_all_branches_with_while(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        value = -1\n'
            '    else:\n'
            '        while False:\n'
            '            value = 1\n'
            '    assert value == -2\n',
            '  File "", line 7\n'
            '        assert value == -2\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_value_in_all_branches_with_for(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        value = -1\n'
            '    else:\n'
            '        for value in range(2):\n'
            '            value += 1\n'
            '    assert value == -2\n',
            '  File "", line 7\n'
            '        assert value == -2\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_value_in_all_branches_with_for_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        value = -1\n'
            '    else:\n'
            '        for _ in range(2):\n'
            '            value = 1\n'
            '    assert value == -2\n',
            '  File "", line 7\n'
            '        assert value == -2\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_value_in_all_branches_with_while_raises(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        value = -1\n'
            '    else:\n'
            '        while True:\n'
            '            raise ValueError()\n'
            '    assert value == -2\n',
            '  File "", line 7\n'
            '        assert value == -2\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_value_in_all_branches_with_for_raises(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if False:\n'
            '        value = -1\n'
            '    else:\n'
            '        for _ in range(1):\n'
            '            raise ValueError()\n'
            '    assert value == -2\n',
            '  File "", line 7\n'
            '        assert value == -2\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Miscellaneous stuff for coverage.py."""

import errno
import hashlib
import inspect
import os
import os.path
import random
import socket


def join_regex(regexes):
    """Combine a list of regexes into one that matches any of them."""
    return "|".join("(?:%s)" % r for r in regexes)


def file_be_gone(path):
    """Remove a file, and don't get annoyed if it doesn't exist."""
    try:
        os.remove(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def ensure_dir(directory):
    """Make sure the directory exists.

    If `directory` is None or empty, do nothing.
    """
    if directory and not os.path.isdir(directory):
        os.makedirs(directory)


def ensure_dir_for_file(path):
    """Make sure the directory for the path exists."""
    ensure_dir(os.path.dirname(path))


def filename_suffix(suffix):
    """Compute a filename suffix for a data file.

    If `suffix` is a string or None, simply return it. If `suffix` is True,
    then build a suffix incorporating the hostname, process id, and a random
    number.

    Returns a string or None.

    """
    if suffix is True:
        # If data_suffix was a simple true value, then make a suffix with
        # plenty of distinguishing information.  We do this here in
        # `save()` at the last minute so that the pid will be correct even
        # if the process forks.
        dice = random.Random(os.urandom(8)).randint(0, 999999)
        suffix = "%s.%s.%06d" % (socket.gethostname(), os.getpid(), dice)
    return suffix


class Hasher:
    """Hashes Python data into md5."""
    def __init__(self):
        self.md5 = hashlib.md5()

    def update(self, v):
        """Add `v` to the hash, recursively if needed."""
        self.md5.update(str(type(v)).encode('utf-8'))
        if isinstance(v, str):
            self.md5.update(v.encode('utf8'))
        elif isinstance(v, bytes):
            self.md5.update(v)
        elif v is None:
            pass
        elif isinstance(v, (int, float)):
            self.md5.update(str(v).encode('utf-8'))
        elif isinstance(v, (tuple, list)):
            for e in v:
                self.update(e)
        elif isinstance(v, dict):
            keys = v.keys()
            for k in sorted(keys):
                self.update(k)
                self.update(v[k])
        else:
            for k in dir(v):
                if k.startswith('__'):
                    continue
                aa = getattr(v, k)
                if inspect.isroutine(aa):
                    continue
                self.update(k)
                self.update(aa)
        self.md5.update(b'.')

    def hexdigest(self):
        """Retrieve the hex digest of the hash."""
        return self.md5.hexdigest()


class DefaultValue:
    """A sentinel object to use for unusual default-value needs.

    Construct with a string that will be used as the repr, for display in help
    and Sphinx output.

    """
    def __init__(self, display_as):
        self.display_as = display_as

    def __repr__(self):
        return self.display_as


class BaseCoverageException(Exception):
    """The base of all Coverage exceptions."""


class CoverageException(BaseCoverageException):
    """An exception raised by a coverage.py function."""


class NoSource(CoverageException):
    """We couldn't find the source for a module."""



from .utils import TestCase
from .utils import build_and_test_module
from .utils import transpile_source


class Test(TestCase):

    def test_char(self):
        build_and_test_module('variables')

    def test_undefined_variable_1(self):
        # Everything ok, 'value' defined.
        transpile_source('def foo(value: bool) -> bool:\n'
                         '    return value\n')

        # Error, 'value' is not defined.
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return value\n',
            '  File "", line 2\n'
            '        return value\n'
            '               ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_undefined_variable_2(self):
        self.assert_transpile_raises(
            'def foo() -> i32:\n'
            '    return 2 * value\n',
            '  File "", line 2\n'
            '        return 2 * value\n'
            '                   ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_undefined_variable_3(self):
        self.assert_transpile_raises(
            'def foo(v1: i32, v2: i32) -> i32:\n'
            '    return v1 + v2\n'
            'def bar() -> i32:\n'
            '    a: i32 = 1\n'
            '    return foo(a, value)\n',
            '  File "", line 5\n'
            '        return foo(a, value)\n'
            '                      ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_undefined_variable_5(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: i8 = a\n',
            '  File "", line 2\n'
            "        a: i8 = a\n"
            '                ^\n'
            "CompileError: undefined variable 'a'\n")

    def test_undefined_variable_6(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: [i8] = a\n',
            '  File "", line 2\n'
            "        a: [i8] = a\n"
            '                  ^\n'
            "CompileError: undefined variable 'a'\n")

    def test_undefined_variable_7(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if a == "":\n'
            '        print("hej")\n',
            '  File "", line 2\n'
            '        if a == "":\n'
            '           ^\n'
            "CompileError: undefined variable 'a'\n")

    def test_undefined_variable_in_fstring(self):
        self.assert_transpile_raises(
            'def bar():\n'
            '    print(f"{value}")\n',
            '  File "", line 2\n'
            '        print(f"{value}")\n'
            '                 ^\n'
            "CompileError: undefined variable 'value'\n")

    def test_undefined_variable_index(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    bar[0] = True\n',
            '  File "", line 2\n'
            '        bar[0] = True\n'
            '        ^\n'
            "CompileError: undefined variable 'bar'\n")

    def test_only_global_defined_in_callee(self):
        self.assert_transpile_raises(
            'GLOB: bool = True\n'
            'def bar() -> i32:\n'
            '    a: i32 = 1\n'
            '    return a\n'
            'def foo() -> i32:\n'
            '    return GLOB + a\n',
            '  File "", line 6\n'
            '        return GLOB + a\n'
            '                      ^\n'
            "CompileError: undefined variable 'a'\n")

    def test_non_snake_case_global_variable(self):
        self.assert_transpile_raises(
            'Aa: i32 = 1\n',
            '  File "", line 1\n'
            '    Aa: i32 = 1\n'
            '    ^\n'
            "CompileError: global variable names must be upper case snake case\n")

    def test_non_snake_case_local_variable(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    A: i32 = 1\n',
            '  File "", line 2\n'
            '        A: i32 = 1\n'
            '        ^\n'
            "CompileError: local variable names must be snake case\n")

    def test_underscore_variable_name(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    _: i32 = 1\n',
            '  File "", line 2\n'
            '        _: i32 = 1\n'
            '        ^\n'
            "CompileError: local variable names must be snake case\n")

    def test_underscore_inferred_variable_name(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    _ = 1\n',
            '  File "", line 2\n'
            '        _ = 1\n'
            '        ^\n'
            "CompileError: local variable names must be snake case\n")

    def test_non_snake_case_local_inferred_variable(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    A = 1\n',
            '  File "", line 2\n'
            '        A = 1\n'
            '        ^\n'
            "CompileError: local variable names must be snake case\n")

    def test_global_variables_can_not_be_redefeined(self):
        self.assert_transpile_raises(
            'A: u8 = 1\n'
            'A: u8 = 2\n',
            '  File "", line 2\n'
            '    A: u8 = 2\n'
            '    ^\n'
            "CompileError: there is already a variable called 'A'\n")

    def test_global_variable_types_can_not_be_inferred(self):
        self.assert_transpile_raises(
            'a = 2\n',
            '  File "", line 1\n'
            '    a = 2\n'
            '    ^\n'
            "CompileError: global variable types cannot be inferred\n")

    def test_no_variable_init(self):
        self.assert_transpile_raises(
            'def foo():\n'
            "    a: u8\n"
            '    a = 1\n'
            '    print(a)\n',
            '  File "", line 2\n'
            "        a: u8\n"
            '        ^\n'
            "CompileError: variables must be initialized when declared\n")

    def test_global_undefined_variable(self):
        self.assert_transpile_raises(
            'V: i8 = (1 << K)\n',
            '  File "", line 1\n'
            '    V: i8 = (1 << K)\n'
            '                  ^\n'
            "CompileError: undefined variable 'K'\n")

    def test_global_use_variable_of_wrong_type(self):
        self.assert_transpile_raises(
            'K: u8 = 1\n'
            'V: i8 = (1 << K)\n',
            '  File "", line 2\n'
            '    V: i8 = (1 << K)\n'
            '             ^\n'
            "CompileError: expected a 'i8', got a 'u8'\n")

    def test_global_integer(self):
        self.assert_transpile_raises(
            '1\n',
            '  File "", line 1\n'
            '    1\n'
            '    ^\n'
            "CompileError: syntax error\n")

    def test_unknown_local_variable_type(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: u9 = 0\n',
            '  File "", line 2\n'
            '        a: u9 = 0\n'
            '           ^\n'
            "CompileError: undefined type 'u9'\n")

    def test_unknown_global_variable_type_1(self):
        self.assert_transpile_raises(
            'A: i9 = 0\n',
            '  File "", line 1\n'
            '    A: i9 = 0\n'
            '       ^\n'
            "CompileError: undefined type 'i9'\n")

    def test_unknown_global_variable_type_2(self):
        self.assert_transpile_raises(
            'A: [(bool, i9)] = None\n',
            '  File "", line 1\n'
            '    A: [(bool, i9)] = None\n'
            '       ^\n'
            "CompileError: undefined type '[(bool, i9)]'\n")

    def test_unknown_global_variable_type_3(self):
        self.assert_transpile_raises(
            'A: i10 = a\n',
            '  File "", line 1\n'
            '    A: i10 = a\n'
            '       ^\n'
            "CompileError: undefined type 'i10'\n")



from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_coverage(self):
        build_and_test_module('coverage', ['--coverage'])

        with open('tests/build/test_coverage/.mys-coverage.txt', 'r') as fin:
            mys_coverage = fin.read()

        self.assert_in(
            'File: ./src/coverage.mys\n'
            '1 0\n'
            '2 0\n'
            '4 3\n'
            '5 3\n'
            '7 3\n'
            '8 1\n'
            '9 2\n'
            '11 0\n'
            '14 2\n'
            '15 2\n'
            '17 3\n'
            '18 0\n'
            '20 3\n'
            '21 6\n'
            '22 3\n'
            '23 3\n'
            '24 0\n'
            '26 3\n'
            '28 2\n'
            '29 2\n'
            '30 1\n'
            '31 1\n'
            '32 0\n'
            '33 0\n'
            '34 1\n'
            '35 1\n'
            '43 1\n'
            '44 1\n'
            '49 1\n'
            '50 1\n'
            '52 0\n'
            '53 0\n'
            '80 1\n'
            '81 1\n'
            '83 1\n'
            '88 1\n'
            '94 1\n'
            '96 1\n'
            '109 1\n'
            '110 1\n'
            '112 0\n'
            '113 0\n'
            '117 1\n'
            '118 1\n'
            '125 1\n'
            '126 1\n'
            '127 1\n'
            '128 0\n'
            '129 0\n',
            mys_coverage)

        self.assert_file_not_exists('tests/build/test_coverage/build/debug/test')
        self.assert_file_exists('tests/build/test_coverage/build/debug-coverage/test')
        self.assert_file_exists('tests/build/test_coverage/coverage/html/index.html')
        self.assert_file_exists('tests/build/test_coverage/.coverage')
        self.assert_file_exists('tests/build/test_coverage/.mys-coverage.txt')

        with Path('tests/build/test_coverage'):
            with patch('sys.argv', ['mys', 'clean']):
                mys.cli.main()

        self.assert_file_not_exists('tests/build/test_coverage/debug-coverage')
        self.assert_file_not_exists('tests/build/test_coverage/.coverage')
        self.assert_file_not_exists('tests/build/test_coverage/.mys-coverage.txt')



import os
import shutil
import subprocess
from io import StringIO
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import create_new_package
from .utils import remove_ansi
from .utils import remove_build_directory
from .utils import run_mys_command


class Test(TestCase):

    def test_foo_new_and_run(self):
        package_name = 'test_foo_new_and_run'
        remove_build_directory(package_name)

        with Path('tests/build'):
            command = [
                'mys', 'new',
                '--author', 'Test Er <test.er@mys.com>',
                package_name
            ]

            stdout = StringIO()

            with patch('sys.stdout', stdout):
                with patch('sys.argv', command):
                    mys.cli.main()

        self.assert_in(
            '┌────────────────────────────────────────────────── 💡 ─┐\n'
            '│ Build and run the new package by typing:              │\n'
            '│                                                       │\n'
            f'│ cd {package_name}                               │\n'
            '│ mys run                                               │\n'
            '└───────────────────────────────────────────────────────┘\n',
            remove_ansi(stdout.getvalue()))

        self.assert_files_equal(f'tests/build/{package_name}/package.toml',
                                'tests/files/foo/package.toml')
        self.assert_files_equal(f'tests/build/{package_name}/.gitignore',
                                'tests/files/foo/.gitignore')
        self.assert_files_equal(f'tests/build/{package_name}/.gitattributes',
                                'tests/files/foo/.gitattributes')
        self.assert_files_equal(f'tests/build/{package_name}/README.rst',
                                'tests/files/foo/README.rst')
        self.assert_files_equal(f'tests/build/{package_name}/LICENSE',
                                'tests/files/foo/LICENSE')
        self.assert_files_equal(f'tests/build/{package_name}/src/main.mys',
                                'tests/files/foo/src/main.mys')
        self.assert_files_equal(f'tests/build/{package_name}/src/lib.mys',
                                'tests/files/foo/src/lib.mys')
        self.assert_files_equal(f'tests/build/{package_name}/doc/index.rst',
                                'tests/files/foo/doc/index.rst')

        with Path(f'tests/build/{package_name}'):
            # Run.
            self.assertFalse(os.path.exists('./build/default/app'))

            with patch('sys.argv', ['mys', 'run', '-j', '1']):
                mys.cli.main()

            self.assert_file_exists(
                f'build/speed/cpp/src/{package_name}/main.mys.cpp')
            self.assert_file_exists('build/speed/app')

            # Clean.
            self.assert_file_exists('build')

            with patch('sys.argv', ['mys', '-d', 'clean']):
                mys.cli.main()

            self.assert_file_not_exists('build')

            with patch('sys.argv', ['mys', '-d', 'clean']):
                mys.cli.main()

            # Build.
            with patch('sys.argv', ['mys', '-d', 'build', '-j', '1']):
                mys.cli.main()

            self.assert_file_exists('./build/speed/app')

            # Run again, but with run() mock to verify that the
            # application is run.
            run_result = Mock()
            run_mock = Mock(side_effect=run_result)

            with patch('subprocess.run', run_mock):
                with patch('sys.argv', ['mys', 'run', '-j', '1']):
                    mys.cli.main()

            # -j 1 is not passed to make if a jobserver is already
            # running.
            self.assertIn(
                run_mock.mock_calls,
                [
                    [
                        call(['make', '-f', 'build/speed/Makefile', 'all',
                              '-s', 'APPLICATION=yes'],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             encoding='utf-8',
                             close_fds=False,
                             env=None),
                        call(['./build/speed/app'], check=True)
                    ],
                    [
                        call(['make', '-f', 'build/speed/Makefile', 'all',
                              '-j', '1', '-s', 'APPLICATION=yes'],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             encoding='utf-8',
                             close_fds=False,
                             env=None),
                        call(['./build/speed/app'], check=True)
                    ]
                ])

            # Test.
            with patch('sys.argv', ['mys', '-d', 'test', '-j', '1']):
                mys.cli.main()

            self.assert_file_exists('./build/debug/test')

            # Doc.
            with patch('sys.argv', ['mys', '-d', 'doc']):
                mys.cli.main()

            self.assert_file_exists('./build/doc/html/index.html')

    def test_new_author_from_git(self):
        package_name = 'test_new_author_from_git'
        remove_build_directory(package_name)

        check_output_mock = Mock(side_effect=['First Last', 'first.last@test.org'])

        with Path('tests/build'):
            with patch('subprocess.check_output', check_output_mock):
                with patch('sys.argv', ['mys', '-d', 'new', package_name]):
                    mys.cli.main()

        self.assertEqual(
            check_output_mock.mock_calls,
            [
                call(['git', 'config', '--get', 'user.name'], encoding='utf-8'),
                call(['git', 'config', '--get', 'user.email'], encoding='utf-8')
            ])

        self.assert_files_equal(f'tests/build/{package_name}/package.toml',
                                'tests/files/test_new_author_from_git/package.toml')

    def test_new_git_command_failure(self):
        package_name = 'test_new_git_command_failure'
        remove_build_directory(package_name)

        check_output_mock = Mock(side_effect=Exception())
        getuser_mock = Mock(side_effect=['mystester'])

        with Path('tests/build'):
            with patch('subprocess.check_output', check_output_mock):
                with patch('getpass.getuser', getuser_mock):
                    with patch('sys.argv', ['mys', '-d', 'new', package_name]):
                        mys.cli.main()

        self.assertEqual(
            check_output_mock.mock_calls,
            [
                call(['git', 'config', '--get', 'user.name'], encoding='utf-8'),
                call(['git', 'config', '--get', 'user.email'], encoding='utf-8')
            ])

        self.assert_files_equal(f'tests/build/{package_name}/package.toml',
                                f'tests/files/test_{package_name}/package.toml')

    def test_new_multiple_authors(self):
        package_name = 'test_new_multiple_authors'
        remove_build_directory(package_name)

        with Path('tests/build'):
            command = [
                'mys', '-d', 'new',
                '--author', 'Test Er <test.er@mys.com>',
                '--author', 'Test2 Er2 <test2.er2@mys.com>',
                package_name
            ]

            with patch('sys.argv', command):
                mys.cli.main()

        self.assert_files_equal(f'tests/build/{package_name}/package.toml',
                                f'tests/files/{package_name}/package.toml')

    def test_publish(self):
        package_name = 'test_publish'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            post_result = Mock()
            post_result.status_code = 200
            post_result.text = None
            post_mock = Mock(side_effect=[post_result])

            with patch('sys.argv', ['mys', '-d', 'publish']):
                with patch('requests.post', post_mock):
                    mys.cli.main()

            the_call = post_mock.call_args_list[0]
            self.assertEqual(the_call.args[0],
                             'https://mys-lang.org/package/test_publish-0.1.0.tar.gz')

    def test_build_outside_package(self):
        # Empty directory.
        package_name = 'test_build_outside_package'
        remove_build_directory(package_name)

        with Path('tests/build'):
            os.makedirs(package_name)

        with Path(f'tests/build/{package_name}'):
            # Build.
            stdout = StringIO()

            with patch('sys.stdout', stdout):
                with self.assertRaises(SystemExit):
                    with patch('sys.argv', ['mys', 'build', '-j', '1']):
                        mys.cli.main()

                expected = '''\
┌──────────────────────────────────────────────────────────────── 💡 ─┐
│ Current directory does not contain a Mys package (package.toml does │
│ not exist).                                                         │
│                                                                     │
│ Please enter a Mys package directory, and try again.                │
│                                                                     │
│ You can create a new package with mys new <name>.                   │
└─────────────────────────────────────────────────────────────────────┘
'''

            self.assert_in(expected, remove_ansi(stdout.getvalue()))

    def test_verbose_build_and_run(self):
        # New.
        package_name = 'test_verbose_build_and_run'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            # Build.
            stdout = StringIO()

            with patch('sys.stdout', stdout):
                with patch('sys.argv', ['mys', '-d', 'build', '--verbose']):
                    mys.cli.main()

            self.assert_in(
                '✔ Building (',
                remove_ansi(stdout.getvalue()))

            # Run.
            stdout = StringIO()

            with patch('sys.stdout', stdout):
                with patch('sys.argv', ['mys', '-d', 'run', '--verbose']):
                    mys.cli.main()

            self.assert_in(
                '✔ Building (',
                remove_ansi(stdout.getvalue()))

    def test_build_empty_package_should_fail(self):
        package_name = 'test_build_empty_package_should_fail'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            os.remove('src/lib.mys')
            os.remove('src/main.mys')

            # Build.
            stdout = StringIO()

            with patch('sys.stdout', stdout):
                with self.assertRaises(SystemExit):
                    with patch('sys.argv', ['mys', 'build']):
                        mys.cli.main()

            self.assert_in(
                '┌─────────────────────────────────────────────────── ❌️ ─┐\n'
                "│ 'src/' is empty. Please create one or more .mys-files. │\n"
                '└────────────────────────────────────────────────────────┘\n',
                remove_ansi(stdout.getvalue()))

    def test_install_local_package_with_local_dependencies(self):
        package_name = 'test_install_local_package_with_local_dependencies'
        remove_build_directory(package_name)
        shutil.copytree('tests/files/install', f'tests/build/{package_name}')

        with Path(f'tests/build/{package_name}/foo'):
            command = ['mys', '-d', 'install', '--root', '..']

            with patch('sys.argv', command):
                mys.cli.main()

            proc = subprocess.run(['../bin/foo'],
                                  check=True,
                                  capture_output=True,
                                  text=True)
            self.assertEqual(proc.stdout, "hello\n")

    def test_install_package_from_registry(self):
        package_name = 'test_install_package_from_registry'
        remove_build_directory(package_name)
        os.makedirs(f'tests/build/{package_name}')

        with Path(f'tests/build/{package_name}'):
            command = ['mys', 'install', '--root', '.', 'hello_world']
            path = os.getcwd()

            try:
                with patch('sys.argv', command):
                    mys.cli.main()
            finally:
                os.chdir(path)

            proc = subprocess.run(['bin/hello_world'],
                                  check=True,
                                  capture_output=True,
                                  text=True)
            self.assertEqual(proc.stdout, "Hello, world!\n")

    def test_make_jobserver_unavaiable_warning(self):
        package_name = 'test_make_jobserver_unavaiable_warning'
        remove_build_directory(package_name)
        create_new_package(package_name)
        path = os.getcwd()

        with Path(f'tests/build/{package_name}'):
            _, stderr = run_mys_command(['build', '-v'], path)

            self.assertNotIn('jobserver unavailable', stderr)

        with Path(f'tests/build/{package_name}'):
            _, stderr = run_mys_command(['run', '-v'], path)

            self.assertNotIn('jobserver unavailable', stderr)

        with Path(f'tests/build/{package_name}'):
            _, stderr = run_mys_command(['test', '-v'], path)

            self.assertNotIn('jobserver unavailable', stderr)

    def test_error_if_package_name_is_not_snake_case(self):
        name = 'test_error_if_package_name_is_not_snake_case'
        remove_build_directory(name)
        shutil.copytree('tests/files/error_if_package_name_is_not_snake_case',
                        f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            with self.assertRaises(SystemExit) as cm:
                with patch('sys.argv', ['mys', 'build']):
                    mys.cli.main()

        self.assertEqual(str(cm.exception),
                         "package name must be snake case, got 'a-hypen'")



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Reporter foundation for coverage.py."""

from .files import FnmatchMatcher
from .files import prep_patterns
from .misc import CoverageException


def get_analysis_to_report(coverage, morfs):
    """Get the files to report on.

    For each morf in `morfs`, if it should be reported on (based on the omit
    and include configuration options), yield a pair, the `FileReporter` and
    `Analysis` for the morf.

    """
    file_reporters = coverage._get_file_reporters(morfs)
    config = coverage.config

    if config.report_include:
        matcher = FnmatchMatcher(prep_patterns(config.report_include))
        file_reporters = [fr for fr in file_reporters if matcher.match(fr.filename)]

    if config.report_omit:
        matcher = FnmatchMatcher(prep_patterns(config.report_omit))
        file_reporters = [fr
                          for fr in file_reporters
                          if not matcher.match(fr.filename)]

    if not file_reporters:
        raise CoverageException("No data to report.")

    for fr in sorted(file_reporters):
        analysis = coverage._analyze(fr)
        yield (fr, analysis)



import textwrap

from ..parser import ast
from .base import BaseVisitor
from .body_check_visitor import BodyCheckVisitor
from .context import Context
from .generics import TypeVisitor
from .generics import add_generic_class
from .generics import format_parameters
from .return_checker_visitor import ReturnCheckerVisitor
from .utils import BUILTIN_ERRORS
from .utils import CompileError
from .utils import GenericType
from .utils import InternalError
from .utils import format_default
from .utils import format_method_name
from .utils import format_return_type
from .utils import get_import_from_info
from .utils import has_docstring
from .utils import indent
from .utils import is_private
from .utils import mys_to_cpp_type
from .utils import mys_to_cpp_type_param
from .utils import split_dict_mys_type
from .value_check_type_visitor import ValueCheckTypeVisitor


def create_coverage_exit(module_path, coverage_variables):
    code = [
        '#if defined(MYS_COVERAGE)',
        f'    mys::mys_coverage_file << "File: {module_path}" << "\\n";'
    ]

    for lineno, variable in coverage_variables:
        code.append(
            f'    mys::mys_coverage_file << {lineno} << " " << {variable} << "\\n";')

    code.append('#endif')

    return code


def create_enum_from_integer(enum):
    code = [
        f'{enum.type} {enum.name}_from_value({enum.type} value)',
        '{',
        '    switch (value) {'
    ]

    for name, value in enum.members:
        code += [
            f'    case {value}:',
            f'        return ({enum.type}){enum.name}::{name};'
        ]

    code += [
        '    default:',
        '        mys::make_shared<ValueError>('
        f'mys::String("enum {enum.name} does not contain ") + '
        'mys::String(value))->__throw();',
        '    }',
        '}'
    ]

    return code


class SourceVisitor(ast.NodeVisitor):
    """The source visitor generates C++ code from given AST.

    """


    def __init__(self,
                 namespace,
                 module_levels,
                 module_hpp,
                 filename,
                 source_lines,
                 definitions,
                 module_definitions,
                 skip_tests,
                 specialized_functions,
                 specialized_classes,
                 coverage_variables):
        self.module_levels = module_levels
        self.module_hpp = module_hpp
        self.filename = filename
        self.skip_tests = skip_tests
        self.namespace = namespace
        self.add_package_main = False
        self.before_namespace = []
        self.in_namespace = []
        self.context = Context(module_levels,
                               specialized_functions,
                               specialized_classes,
                               source_lines)
        self.definitions = definitions
        self.module_definitions = module_definitions
        self.enums = []
        self.body = []
        self.variables = []
        self.init_globals = []
        self.method_comprehensions = self.context.method_comprehensions
        self.coverage_exit = create_coverage_exit(filename, coverage_variables)

        for name, functions in module_definitions.functions.items():
            self.context.define_function(
                name,
                self.context.make_full_name_this_module(name),
                functions)

        for name, trait_definitions in module_definitions.traits.items():
            self.context.define_trait(name,
                                      self.context.make_full_name_this_module(name),
                                      trait_definitions)

        self.context.define_trait('Error', 'Error', None)

        for name, class_definitions in module_definitions.classes.items():
            self.context.define_class(name,
                                      self.context.make_full_name_this_module(name),
                                      class_definitions)

        for name in BUILTIN_ERRORS:
            self.context.define_class(name, name, None)

        for enum in module_definitions.enums.values():
            self.context.define_enum(
                enum.name,
                self.context.make_full_name_this_module(enum.name),
                enum)
            self.enums += create_enum_from_integer(enum)

        for name, variable_definitions in module_definitions.variables.items():
            TypeVisitor(self.context).visit(variable_definitions.node.annotation)

    def define_parameters(self, args):
        for param, node in args:
            if isinstance(param.type, GenericType):
                param_type = add_generic_class(param.type.node, self.context)[1]
            else:
                param_type = param.type

            self.raise_if_type_not_defined(param_type, param.node.annotation)
            self.context.define_local_variable(param.name, param_type, node)

    def add_application_init(self, ordered_modules):
        body = []

        for module in ordered_modules:
            self.before_namespace += [
                f'namespace mys::{module.replace(".", "::")} {{',
                'extern void __module_init();',
                '}'
            ]
            body.append(f'    mys::{module.replace(".", "::")}::__module_init();')

        self.before_namespace += [
            'void __application_init(void)',
            '{'
        ] + body + [
            '}'
        ]

    def add_application_exit(self, ordered_modules):
        body = [
            '#if defined(MYS_COVERAGE)',
            '    mys::mys_coverage_file.open(".mys-coverage.txt");',
            '#endif'
        ]

        for module in ordered_modules:
            self.before_namespace += [
                f'namespace mys::{module.replace(".", "::")} {{',
                'extern void __module_exit();',
                '}'
            ]
            body.append(f'    mys::{module.replace(".", "::")}::__module_exit();')

        body += [
            '#if defined(MYS_COVERAGE)',
            '    mys::mys_coverage_file.close();',
            '#endif'
        ]

        self.before_namespace += [
            'void __application_exit(void)',
            '{'
        ] + body + [
            '}'
        ]

    def raise_if_type_not_defined(self, mys_type, node):
        if not self.context.is_type_defined(mys_type):
            raise CompileError(f"undefined type '{mys_type}'", node)

    def visit_AnnAssign(self, node):
        return GlobalVariableVisitor(self.context, '').visit(node)

    def visit_variable(self, variable):
        mys_type = TypeVisitor(self.context).visit(variable.node.annotation)
        cpp_type = mys_to_cpp_type(mys_type, self.context)

        return [f'{cpp_type} {variable.name};']

    def visit_global_variable(self, node, body, next_index):
        docstring = None

        if len(body) > next_index:
            docstring_node = body[next_index]

            if isinstance(docstring_node, ast.Expr):
                if isinstance(docstring_node.value, ast.Constant):
                    if isinstance(docstring_node.value.value, str):
                        docstring = docstring_node.value.value

        self.init_globals += GlobalVariableVisitor(self.context, '').visit(node)

        return docstring is not None

    def visit_Module(self, node):
        body_iter = iter(node.body)
        item_index = 0

        for item in body_iter:
            if isinstance(item, ast.AnnAssign):
                if self.visit_global_variable(item, node.body, item_index + 1):
                    next(body_iter)
                    item_index += 1
            else:
                self.init_globals += self.visit(item)

            item_index += 1

        for name, definitions in self.module_definitions.classes.items():
            if definitions.is_generic():
                continue

            self.body += self.visit_class_definition(name, definitions)

        for functions in self.module_definitions.functions.values():
            for function in functions:
                if function.is_generic():
                    continue

                self.body += self.visit_function_defaults(function)
                self.body += self.visit_function_definition(function)

        for variable in self.module_definitions.variables.values():
            self.variables += self.visit_variable(variable)

    def visit_specialized_function(self, function):
        self.body += self.visit_function_defaults(function)
        self.body += self.visit_function_definition(function)

    def visit_specialized_class(self, name, definitions):
        self.context.define_class(name,
                                  self.context.make_full_name_this_module(name),
                                  definitions)
        self.body += self.visit_class_definition(name, definitions)

    def format_cpp(self):
        traceback_entries = ',\n'.join([
            f'    {{ "{name}", {lineno}, "{code}" }}'
            for name, lineno, code in self.context.traceback.entries
        ])
        return '\n'.join([
            '// This file was generated by mys. DO NOT EDIT!!!',
            '#include "mys.hpp"',
            f'#include "{self.module_hpp}"'
        ] + self.before_namespace + [
            f'namespace {self.namespace}',
            '{',
            'static mys::TracebackEntryInfo __traceback_entries_info[] = {',
            traceback_entries,
            '};',
            'static mys::TracebackModuleInfo __traceback_module_info = {',
            f'   .path_p = "{self.filename}",',
            '   .entries_info_p = &__traceback_entries_info[0]',
            '};'
        ] + self.in_namespace
          + self.enums
          + [constant[1] for constant in self.context.constants.values()]
          + self.context.comprehensions
          + self.variables
          + self.body + [
              'void __module_init()',
              '{'
          ] + self.init_globals + [
              '}',
              'void __module_exit()',
              '{'
          ] + self.coverage_exit + [
              '}',
              '}'
          ] + self.main())

    def main(self):
        if self.add_package_main:
            return [
                'void package_main(int argc, const char *argv[])',
                '{',
                f'    {self.namespace}::main(argc, argv);',
                '}'
            ]
        else:
            return []

    def define_implicitly_imported_types(self, mys_type):
        if isinstance(mys_type, tuple):
            for item_mys_type in mys_type:
                self.define_implicitly_imported_types(item_mys_type)
        elif isinstance(mys_type, list):
            self.define_implicitly_imported_types(mys_type[0])
        elif isinstance(mys_type, dict):
            key_mys_type, value_mys_type = split_dict_mys_type(mys_type)
            self.define_implicitly_imported_types(key_mys_type)
            self.define_implicitly_imported_types(value_mys_type)
        elif isinstance(mys_type, GenericType):
            # ToDo, but what should be done?
            pass
        elif '.' in mys_type:
            module = '.'.join(mys_type.split('.')[:-1])
            name = mys_type.split('.')[-1]
            imported = self.definitions.get(module)

            if imported is None:
                raise Exception(f"imported module '{module}' not found")
            elif name in imported.classes:
                self.context.define_class(mys_type, mys_type, imported.classes[name])
            elif name in imported.traits:
                self.context.define_trait(mys_type, mys_type, imported.traits[name])

    def define_imported_function_parameters_and_return_type(self, function):
        if function.returns is not None:
            self.define_implicitly_imported_types(function.returns)

        for param, _ in function.args:
            self.define_implicitly_imported_types(param.type)

    def define_imported_class_member(self, member):
        self.define_implicitly_imported_types(member.type)

    def visit_ImportFrom(self, node):
        module, name, asname = get_import_from_info(node, self.module_levels)
        imported_module = self.definitions.get(module)

        if imported_module is None:
            raise CompileError(f"imported module '{module}' does not exist", node)

        if is_private(name):
            raise CompileError(f"cannot import private definition '{name}'", node)

        full_name = f'{module}.{name}'

        if name in imported_module.variables:
            self.context.define_global_variable(
                asname,
                full_name,
                imported_module.variables[name].type,
                node)
        elif name in imported_module.functions:
            for function in imported_module.functions[name]:
                self.define_imported_function_parameters_and_return_type(function)

            self.context.define_function(asname,
                                         full_name,
                                         imported_module.functions[name])
        elif name in imported_module.classes:
            for methods in imported_module.classes[name].methods.values():
                for method in methods:
                    self.define_imported_function_parameters_and_return_type(method)

            for member in imported_module.classes[name].members.values():
                self.define_imported_class_member(member)

            self.context.define_class(asname,
                                      full_name,
                                      imported_module.classes[name])
        elif name in imported_module.traits:
            self.context.define_trait(asname,
                                      full_name,
                                      imported_module.traits[name])
        elif name in imported_module.enums:
            self.context.define_enum(asname,
                                     full_name,
                                     imported_module.enums[name])
        else:
            raise CompileError(
                f"imported module '{module}' does not contain '{name}'",
                node)

        return []

    def visit_ClassDef(self, _node):
        return []

    def visit_defaults(self, name, args):
        code = []

        for param, default in args:
            if default is None:
                continue

            cpp_type = mys_to_cpp_type(param.type, self.context)
            body = ValueCheckTypeVisitor(
                BaseVisitor(self.context, self.filename)).visit(default, param.type)

            code += [
                format_default(name, param.name, cpp_type),
                '{',
                f'    return {body};',
                '}'
            ]

        return code

    def visit_method_defaults(self, method, class_name):
        if method.name == '__init__':
            method_name = class_name
        elif method.name == '__del__':
            method_name = f'~{class_name}'
        else:
            method_name = method.name

        return self.visit_defaults(f'{class_name}_{method_name}', method.args)

    def visit_class_methods_definition(self,
                                       class_name,
                                       method_names,
                                       methods_definitions):
        body = []
        self.context.class_name = class_name

        for method in methods_definitions:
            body += self.visit_method_defaults(method, class_name)
            self.context.push()
            self.context.define_local_variable(
                'self',
                self.context.make_full_name_this_module(class_name),
                method.node.args.args[0])
            self.define_parameters(method.args)
            self.raise_if_type_not_defined(method.returns, method.node.returns)

            if method.returns is not None:
                ReturnCheckerVisitor().visit(method.node)

            method_names.append(method.name)
            method_name = format_method_name(method, class_name)
            parameters = format_parameters(method.args, self.context)
            self.context.return_mys_type = method.returns

            if method_name in [class_name, f'~{class_name}']:
                body.append(f'{class_name}::{method_name}({parameters})')
            else:
                return_cpp_type = format_return_type(method.returns, self.context)
                body.append(
                    f'{return_cpp_type} {class_name}::{method_name}({parameters})')

            body.append('{')
            body.append(self.context.traceback.enter(method.name))
            body_iter = iter(method.node.body)

            if has_docstring(method.node):
                next(body_iter)

            for item in body_iter:
                BodyCheckVisitor().visit(item)
                body.append(self.context.traceback.set(item.lineno))
                body.append(indent(BodyVisitor(self.context,
                                               self.filename).visit(item)))

            body.append(self.context.traceback.exit())
            body.append('}')
            self.context.pop()

        self.context.class_name = None

        return body

    def visit_class_definition(self, class_name, definitions):
        member_cpp_types = []
        member_names = []
        method_names = []
        body = []

        for member in definitions.members.values():
            if isinstance(member.type, GenericType):
                member_type = add_generic_class(member.type.node, self.context)[1]
            else:
                member_type = member.type

            if not self.context.is_type_defined(member_type):
                raise CompileError(f"undefined type '{member_type}'",
                                   member.node.annotation)

            member_cpp_types.append(mys_to_cpp_type_param(member_type, self.context))
            member_names.append(member.name)

        for methods in definitions.methods.values():
            body += self.visit_class_methods_definition(class_name,
                                                        method_names,
                                                        methods)

        if 'Error' in definitions.implements:
            body += [
                f'void {class_name}::__throw()',
                '{',
                f'    throw __{class_name}(mys::shared_ptr<{class_name}>(this));',
                '}'
            ]

        return body

    def visit_FunctionDef(self, _node):
        return []

    def visit_function_defaults(self, function):
        return self.visit_defaults(function.make_name(), function.args)

    def visit_function_definition_main(self, function, parameters, body):
        self.add_package_main = True

        if parameters not in ['SharedList<mys::String> argv', 'void']:
            raise CompileError("main() takes 'argv: [string]' or no arguments",
                               function.node)

        if function.returns is not None:
            raise CompileError("main() must not return any value", function.node)

        if parameters == 'void':
            body = [
                '    (void)__argc;',
                '    (void)__argv;'
            ] + body
        else:
            body = ['    auto argv = create_args(__argc, __argv);'] + body

        parameters = 'int __argc, const char *__argv[]'

        return body, parameters

    def visit_function_definition_test(self, function, parameters, prototype, body):
        if self.skip_tests:
            return []

        if parameters != 'void':
            raise CompileError("test functions takes no parameters",
                               function.node)

        if function.returns is not None:
            raise CompileError("test functions must not return any value",
                               function.node)

        namespace = '::'.join(self.module_levels[1:])
        code = [
            '#if defined(MYS_TEST)',
            f'static {prototype}',
            '{'
        ] + body + [
            '}',
            f'static Test mys_test_{function.name}("{namespace}::{function.name}", '
            f'{function.name});',
            '#endif'
        ]

        return code

    def visit_function_definition(self, function):
        self.context.push()
        self.define_parameters(function.args)
        self.raise_if_type_not_defined(function.returns, function.node.returns)

        if function.returns is not None:
            ReturnCheckerVisitor().visit(function.node)

        function_name = function.make_name()
        parameters = format_parameters(function.args, self.context)
        return_cpp_type = format_return_type(function.returns, self.context)
        self.context.return_mys_type = function.returns
        body = [self.context.traceback.enter(function.name)]
        body_iter = iter(function.node.body)

        if has_docstring(function.node):
            next(body_iter)

        for item in body_iter:
            BodyCheckVisitor().visit(item)
            body.append(self.context.traceback.set(item.lineno))
            body.append(indent(BodyVisitor(self.context, self.filename).visit(item)))

        body.append(self.context.traceback.exit())

        if function_name == 'main':
            body, parameters = self.visit_function_definition_main(function,
                                                                   parameters,
                                                                   body)

        prototype = f'{return_cpp_type} {function_name}({parameters})'

        if function.is_test:
            code = self.visit_function_definition_test(function,
                                                       parameters,
                                                       prototype,
                                                       body)
        else:
            code = [
                prototype,
                '{'
            ] + body + [
                '}'
            ]

        self.context.pop()

        return code

    def visit_Expr(self, node):
        return self.visit(node.value) + [';']

    def visit_Constant(self, node):
        if isinstance(node.value, tuple) and len(node.value) == 1:
            value = node.value[0]

            if value.startswith('source-before-namespace'):
                self.before_namespace += [
                    '/* c-string-source-before-namespace start */',
                    textwrap.dedent(value[23:]).strip(),
                    '/* c-string-source-before-namespace stop */'
                ]
            elif value.startswith('header-before-namespace'):
                pass
            else:
                self.in_namespace += [
                    '/* c-string start */',
                    textwrap.dedent(value).strip(),
                    '/* c-string stop */']

            return []

        raise CompileError("syntax error", node)

    def generic_visit(self, node):
        raise InternalError("unhandled node", node)


class BodyVisitor(BaseVisitor):
    pass


class GlobalVariableVisitor(BaseVisitor):

    def visit_AnnAssign(self, node):
        target, mys_type, _, cpp_target, code = self.visit_ann_assign(node)
        self.context.define_global_variable(
            target,
            self.context.make_full_name_this_module(target),
            mys_type,
            node.target)

        return [f'    {cpp_target} = {code};']
